import contextlib
from functools import lru_cache
import json
import os
import os.path as op
from pathlib import Path
import random
import sys
import time

import click
import girder_client as gcl
from keyring.backend import get_all_keyring
from keyring.core import get_keyring, load_config, load_env
from keyring.errors import KeyringError
from keyring.util.platform_ import config_root
from keyrings.alt.file import EncryptedKeyring

from .consts import MAX_CHUNK_SIZE, known_instances_rev
from .exceptions import LockingError
from . import get_logger
from .utils import ensure_datetime, flatten, flattened, remap_dict, try_multiple

lgr = get_logger()


# TODO: more flexible
class GirderServer:
    __slots__ = ["url"]

    def __init__(self, url):
        self.url = url


class GirderNotFound(Exception):
    pass


# TODO: remove if we start to expose this as a Python library which could
# be longer lived than just a CLI
@lru_cache(1024)
def lookup(client, name, asset_type="collection", path=None):
    """A helper for common logic while looking up things on girder by name"""
    req = f"/{asset_type}/{name}"
    target_msg = asset_type.capitalize()
    if path is not None:
        req += f"/{path}"
        target_msg = "Path"

    try:
        return client.resourceLookup(req)
    except gcl.HttpError as exc:
        response = get_HttpError_response(exc) or {}
        if not (response and f"{target_msg} not found" in response["message"]):
            raise exc  # we dunno much about this - so just reraise
        # but if it was indeed just a miss -- raise our dedicated Exception
        lgr.debug(f"{target_msg} was not found: {response}")
        raise GirderNotFound(response)


def get_HttpError_response(exc):
    """Given an gcl.HttpError exception instance, parse and return response

    Returns
    -------
    None or dict
      If exception does not contain valid json response record, returns None
    """
    try:
        responseText = getattr(exc, "responseText", "")
        return json.loads(responseText)
    except Exception as exc2:
        lgr.debug("Cannot parse response %s as json: %s", responseText, exc2)
    return None


def get_HttpError_message(exc):
    resp = get_HttpError_response(exc)
    if isinstance(resp, dict):
        return resp.get("message", None)
    return resp


def is_access_denied(exc):
    """Tell if an exception about denied access"""
    response = get_HttpError_response(exc)
    return exc.status == 401 or (
        response and "access denied" in response.get("message", "")
    )


# Provide additional "heavy" logging at DEBUG level about interactions
# with girder, including progress indication.  The need came up to troubleshoot
# https://github.com/dandi/dandi-cli/issues/136
_DANDI_LOG_GIRDER = os.environ.get("DANDI_LOG_GIRDER")


class GirderCli(gcl.GirderClient):
    """An "Adapter" to GirderClient

    ATM it just inherits from GirderClient. But in the long run we might want
    to make it fulfill our API and actually become and adapter (and delegate)
    to GirderClient.  TODO
    """

    def __init__(self, server_url, progressbars=False):
        self._server_url = server_url.rstrip("/")
        kw = {}
        if progressbars:
            kw["progressReporterCls"] = TQDMProgressReporter
        super().__init__(
            apiUrl="{}/api/v1".format(self._server_url),
            # seems to mess up our "dandi upload" reporting, although
            # I thought that it is used only for download. heh heh
            **kw,
        )

    if _DANDI_LOG_GIRDER:
        # Overload this core method to be able to log interactions with girder
        # server from the client side
        def sendRestRequest(self, *args, **kwargs):
            lgr.debug("REST>: args=%s kwargs=%s", args, kwargs)
            res = super().sendRestRequest(*args, **kwargs)
            lgr.debug("REST<: %s", str(res))
            return res

    def register_dandiset(self, name, description):
        """Register a dandiset and return created metadata record

        Returns
        -------
        dict
          Metadata record created for the dandiset
        """
        ret = self.createResource("dandi", {"name": name, "description": description})
        assert "meta" in ret
        assert "dandiset" in ret["meta"]
        # Return only our dandiset record
        return ret["meta"]["dandiset"]

    def dandi_authenticate(self):
        # Shortcut for advanced folks
        api_key = os.environ.get("DANDI_GIRDER_API_KEY", None)
        if api_key:
            self.authenticate(apiKey=api_key)
            lgr.debug("Successfully authenticated using the key from the envvar")
            return

        if self._server_url in known_instances_rev:
            client_name = known_instances_rev[self._server_url]
        else:
            raise NotImplementedError("TODO client name derivation for keyring")

        app_id = "dandi-girder-{}".format(client_name)
        keyring_backend, api_key = keyring_lookup(app_id, "key")
        # the dance about rejected authentication etc
        while True:
            if not api_key:
                api_key = input(
                    "Please provide API Key (created/found in My "
                    "Account/API keys "
                    "in Girder) for {}: ".format(client_name)
                )
                keyring_backend.set_password(app_id, "key", api_key)
                lgr.debug("Stored key in keyring")

            try:
                self.authenticate(apiKey=api_key)
                lgr.debug("Successfully authenticated using the key")
                break
            except Exception as exc:
                sys.stderr.write("Failed to authenticate: {}".format(exc))
                api_key = None

    # XXX causes strange 'NoneType' object has no attribute 'close'
    #  so TODO to investigate
    # def sendRestRequest(self, *args, **kwargs):
    #     with self.session() as session:
    #         # TODO: avoid hardcoded max_retries -- setup taken from girder_cli.client
    #         session.mount(self.urlBase, HTTPAdapter(max_retries=5))
    #         return super(GirderCli, self).sendRestRequest(*args, **kwargs)

    @classmethod
    def _adapt_record(cls, rec):
        # we will care about subset of fields girder provides and will reshape
        # them a little
        mapping = {
            "_id": "id",
            "name": "name",
            "_modelType": "type",
            "meta": "metadata",
            "size": ("attrs", "size"),  # only 1 level supported ATM
            "updated": ("attrs", "mtime"),
            "created": ("attrs", "ctime"),
        }
        # TODO: figure out mtime/ctime between item and a file
        # On a sample (uploaded not using dandi-cli) upload -- item has both mtime/ctime
        # and a file -- only ctime
        #  (Pdb) p rec
        # {'_id': '5e60c19381bc3e47d94aa014', '_modelType': 'item',
        # 'baseParentId': '5da4b8fe51c340795cb18fd0', 'baseParentType':
        # 'user', 'created': '2020-03-05T09:08:35.193000+00:00', 'creatorId':
        # '5da4b8fe51c340795cb18fd0', 'description': '', 'folderId':
        # '5e60c14f81bc3e47d94aa012', 'meta': {}, 'name': '18516000.nwb',
        # 'size': 792849, 'updated': '2020-03-05T09:08:35.193000+00:00'}
        # (Pdb) c
        # > /home/yoh/proj/dandi/dandi-cli/dandi/girder.py(136)_adapt_record()
        # -> for girder, dandi in mapping.items():
        # (Pdb) p rec
        # {'_id': '5e60c19381bc3e47d94aa015', '_modelType': 'file',
        # 'created': '2020-03-05T09:08:35.196000+00:00', 'creatorId':
        # '5da4b8fe51c340795cb18fd0', 'downloadStatistics': {'completed': 1,
        # 'requested': 1, 'started': 1}, 'exts': ['nwb'], 'itemId':
        # '5e60c19381bc3e47d94aa014', 'mimeType': 'application/octet-stream',
        # 'name': '18516000.nwb', 'size': 792849}
        rec_out = {}
        for girder, dandi in mapping.items():
            if girder not in rec:
                continue  # skip
            v = rec[girder]
            if isinstance(dandi, (tuple, list)):
                if dandi[0] not in rec_out:
                    rec_out[dandi[0]] = {}
                rec_out[dandi[0]][dandi[1]] = v
            else:
                rec_out[dandi] = v
        return rec_out

    def traverse_asset(self, asset_id, asset_type, parent_path=None, recursive=True):
        """Generator to produce asset records

        ATM the fields I see possibly non-exhaustive and would not be needed
        for some cases.  Later we could make some structure with on-demand
        late requests (e.g. if ["metadata"] is requested, request it at that
        point.

        We will "meld" notion of a Girder Item with File ATM.  File detail
        matters, Item not -- we just would "steal" its

        TODO:
        - make asset_type optional? i.e. discover it since IDs are unique
        - add "include" "exclude" filters to avoid even going into
          anything of no interest?  note that for folder/collection we now
          performing size computation.  So those would get affected, but
          could be taken as a feature.
        - makeup "children_unique_metadata" for folders?

        Parameters
        ----------

        Yields
        ------
         dicts with keys:
           type: str, {file, folder, dandiset}
           id: str, internal to backend id
           path: str, path within the asset
           attrs: dict, expected keys 'size', 'mtime', 'ctime' when known
           metadata: dict
        """
        # TODO: dandiset.  We can request folder id, verify that it has dataset.yaml
        # and populate metadata for the record with that content
        attempts_left = 4
        while attempts_left:
            attempts_left -= 1
            try:
                g = self.getResource(asset_type, asset_id)
                break
            except gcl.HttpError as exc:
                if not self.token and is_access_denied(exc):
                    lgr.warning("unauthenticated access denied, let's authenticate")
                    self.dandi_authenticate()
                else:
                    raise
        if not attempts_left:
            raise RuntimeError("Failed to authenticate after a number of attempts")
        yield from self._traverse_asset_girder(g, parent_path, recursive=recursive)

    def _traverse_asset_girder(self, g, parent_path=None, recursive=True):
        """Helper which operates on girder record"""
        import requests

        a = self._adapt_record(g)
        if parent_path:
            a["path"] = op.join(parent_path, a["name"])
        else:
            a["path"] = a["name"]

        if a["type"] == "file":
            # we are done, but hopefully it never sees outside since
            # item should be the one we care about
            pass
        elif a["type"] == "item":
            file_recs = try_multiple(5, requests.ConnectionError, 1.1)(
                lambda: list(self.listFile(g["_id"]))
            )
            if len(file_recs) > 1:
                lgr.warning("Multiple files found for %s; using oldest one", a["path"])
                file_recs = [
                    min(file_recs, key=lambda fr: ensure_datetime(fr["created"]))
                ]
            if not file_recs:
                lgr.warning(f"Ran into an empty item: {g}")
                return
            # !!!! double trip
            a_file = next(
                self._traverse_asset_girder(file_recs[0], parent_path=parent_path)
            )
            assert not a_file.get("metadata", None), "metadata should be per item"
            a_file["metadata"] = metadata = a["metadata"]
            # and since girder does not allow non-admin set 'ctime' for the file
            # we will use our metadata records
            if "uploaded_mtime" in metadata:
                a_file["attrs"]["mtime"] = metadata["uploaded_mtime"]
            a = a_file  # yield enhanced with metadata file entry
        elif a["type"] in ("folder", "collection", "user"):
            if recursive:
                # TODO: reconsider bothering with types, since could be done
                # upstairs and returning folder itself last complicates reporting
                # of intensions.  We could yield it twice, but then it might bring
                # confusion
                a["attrs"]["size"] = 0
                for child in self._list_folder(g["_id"], g["_modelType"]):
                    for child_a in self._traverse_asset_girder(child, a["path"]):
                        a["attrs"]["size"] += child_a["attrs"]["size"]
                        yield child_a
            # It could be a dandiset
            if a["type"] == "folder" and a["metadata"].get("dandiset", {}).get(
                "identifier", None
            ):
                a["type"] = "dandiset"
            else:
                # And now yield record about myself, but we do not care about collections
                a["type"] = "folder"
        else:
            raise NotImplementedError(f"Do not know how to handle {a['type']!r}")
        yield a

    def _list_folder(self, folder_id, folder_type="folder", types=None):
        """A helper to list a "folder" content

        types: {folder, item}, optional
          Collection cannot contain items, so we automagically decide
        """
        # ... probably could have made use of listResource
        g_id = self._checkResourcePath(folder_id)
        # Different for folder and item
        params = {
            "folder": {"parentType": folder_type, "parentId": g_id},
            "item": {"folderId": folder_id},
        }

        if types is None:
            types = ["folder"]
            if folder_type != "collection":
                types.append("item")

        for child_type in types:
            offset = 0
            while True:
                children = self.get(
                    child_type,
                    parameters=dict(
                        limit=gcl.DEFAULT_PAGE_LIMIT,
                        offset=offset,
                        **params[child_type],
                    ),
                )
                for child in children:
                    yield child

                offset += len(children)
                if len(children) < gcl.DEFAULT_PAGE_LIMIT:
                    break

    def get_download_file_iter(self, file_id, chunk_size=MAX_CHUNK_SIZE):
        def downloader(start_at=0):
            # TODO: make it a common decorator here?
            # Will do 3 attempts to avoid some problems due to flaky/overloaded
            # connections, see https://github.com/dandi/dandi-cli/issues/87
            for attempt in range(3):
                try:
                    path = f"file/{file_id}/download"
                    if start_at > 0:
                        headers = {"Range": f"bytes={start_at}-"}
                        # Range requests result in a 206 response, which the
                        # Girder client treats as an error (at least until they
                        # merge girder/girder#3301).  Hence, we need to make
                        # the request directly through `requests`.
                        import requests

                        resp = requests.get(
                            f"{self._server_url}/api/v1/{path}",
                            stream=True,
                            headers=headers,
                        )
                        resp.raise_for_status()
                    else:
                        resp = self.sendRestRequest(
                            "get", path, stream=True, jsonResp=False
                        )
                    return resp.iter_content(chunk_size=chunk_size)
                except gcl.HttpError as exc:
                    if is_access_denied(exc) or attempt >= 2:
                        raise
                    # sleep a little and retry
                    lgr.debug(
                        "Failed to download on attempt#%d, will sleep a bit and retry",
                        attempt,
                    )
                    time.sleep((1 + random.random()) * 5)

        return downloader

    def _get_asset_recs(self, asset_id, asset_type, authenticate=False, recursive=True):
        """

        Parameters
        ----------
        asset_id
        asset_type
        authenticate
        recursive

        Returns
        -------
        dandiset_rec, files_rec
           dandiset_rec will be None if asset_id is not pointing to a dandiset
        """
        # asset_rec = client.getResource(asset_type, asset_id)
        # lgr.info("Working with asset %s", str(asset_rec))
        # In principle Girder's client already has ability to download any
        # resource (collection/folder/item/file).  But it seems that "mending" it
        # with custom handling (e.g. later adding filtering to skip some files,
        # or add our own behavior on what to do when files exist locally, etc) would
        # not be easy.  So we will reimplement as a two step (kinda) procedure.
        # Return a generator which would be traversing girder and yield records
        # of encountered resources.
        # TODO later:  may be look into making it async
        # First we access top level records just to sense what we are working with
        top_entities = None
        files = None
        dandiset = None

        while True:
            try:
                # this one should enhance them with "fullpath"
                top_entities = list(
                    self.traverse_asset(asset_id, asset_type, recursive=False)
                )
                break
            except gcl.HttpError as exc:
                if not authenticate and is_access_denied(exc):
                    lgr.warning("unauthenticated access denied, let's authenticate")
                    self.dandi_authenticate()
                    continue
                raise
        entity_type = list(set(e["type"] for e in top_entities))
        if len(entity_type) > 1:
            raise ValueError(
                f"Please point to a single type of entity - either dandiset(s),"
                f" folder(s) or file(s).  Got: {entity_type}"
            )
        entity_type = entity_type[0]
        if entity_type in ("dandiset", "folder"):
            # redo recursively
            lgr.info(
                "Traversing remote %ss (%s) recursively",
                entity_type,
                ", ".join(e["name"] for e in top_entities),
            )
            entities = self.traverse_asset(asset_id, asset_type, recursive=recursive)
            # TODO: special handling for a dandiset -- we might need to
            #  generate dandiset.yaml out of the metadata record
            # we care only about files ATM
            files = (e for e in entities if e["type"] == "file")
        elif entity_type == "file":
            files = top_entities
        else:
            raise ValueError(f"Unexpected entity type {entity_type}")
        # TODO: move -- common and has nothing to do with getting a list of assets
        if entity_type == "dandiset":
            if len(top_entities) > 1:
                raise NotImplementedError(
                    "A single dandiset at a time only supported ATM, got %d: %s"
                    % (len(top_entities), top_entities)
                )
            dandiset = top_entities[0]

        return dandiset, files

    def get_dandiset_and_assets(
        self, asset_id, asset_type, recursive=True, authenticate=False
    ):
        lgr.debug(f"Traversing {asset_type} with id {asset_id}")
        # there might be multiple asset_ids, e.g. if multiple files were selected etc,
        # so we will traverse all of them
        dandiset_asset_recs = [
            self._get_asset_recs(
                asset_id_, asset_type, authenticate=authenticate, recursive=recursive
            )
            for asset_id_ in set(flattened([asset_id]))
        ]

        dandiset = None

        if not dandiset_asset_recs:
            lgr.warning("Got empty listing for %s %s", asset_type, asset_id)
            return
        elif (
            len(dandiset_asset_recs) > 1
        ):  # had multiple asset_ids, should not be dandisets
            if any(r[0] for r in dandiset_asset_recs):
                raise NotImplementedError("Got multiple ids for dandisets")
        else:
            dandiset = dandiset_asset_recs[0][0]

        # Return while harmonizing
        if dandiset:
            dandiset = _harmonize_girder_dandiset_to_dandi_api(dandiset)

        return (
            dandiset,
            (
                _harmonize_girder_asset_to_dandi_api(a)
                for a in flatten(r[1] for r in dandiset_asset_recs)
            ),
        )

    @contextlib.contextmanager
    def lock_dandiset(self, dandiset_identifier: str):
        presumably_locked = False
        try:
            lgr.debug("Trying to acquire lock for %s", dandiset_identifier)
            try:
                resp = self.post(f"dandi/{dandiset_identifier}/lock")
                lgr.debug("Locking response: %s", str(resp))
            except gcl.HttpError as exc:
                msg = get_HttpError_message(exc) or str(exc)
                raise LockingError(
                    f"Failed to lock dandiset {dandiset_identifier} due to: {msg}"
                )
            else:
                presumably_locked = True

            yield
        finally:
            if presumably_locked:
                lgr.debug("Trying to release the lock for %s", dandiset_identifier)
                try:
                    resp = self.post(f"dandi/{dandiset_identifier}/unlock")
                    lgr.debug("Unlocking response: %s", str(resp))
                except gcl.HttpError as exc:
                    msg = get_HttpError_message(exc) or str(exc)
                    raise LockingError(
                        f"Failed to unlock dandiset {dandiset_identifier} due to: {msg}"
                    )

    NGINX_MAX_CHUNK_SIZE = 400 * (1 << 20)  # 400 MiB

    def _uploadContents(self, uploadObj, stream, size, progressCallback=None):
        """
        Uploads contents of a file.  Overridden so that the chunk size can be
        set on a per-file basis.

        :param uploadObj: The upload object contain the upload id.
        :type uploadObj: dict
        :param stream: Readable stream object.
        :type stream: file-like
        :param size: The length of the file. This must be exactly equal to the
            total number of bytes that will be read from ``stream``, otherwise
            the upload will fail.
        :type size: str
        :param progressCallback: If passed, will be called after each chunk
            with progress information. It passes a single positional argument
            to the callable which is a dict of information about progress.
        :type progressCallback: callable
        """
        offset = 0
        uploadId = uploadObj["_id"]

        chunk_size = max(self.MAX_CHUNK_SIZE, (size + 999) // 1000)
        if chunk_size > self.NGINX_MAX_CHUNK_SIZE:
            raise Exception("File requires too many chunks to upload")

        with self.progressReporterCls(
            label=uploadObj.get("name", ""), length=size
        ) as reporter:

            while True:
                chunk = stream.read(min(chunk_size, (size - offset)))

                if not chunk:
                    break

                if isinstance(chunk, str):
                    chunk = chunk.encode("utf8")

                uploadObj = self.post(
                    "file/chunk?offset=%d&uploadId=%s" % (offset, uploadId),
                    data=gcl._ProgressBytesIO(chunk, reporter=reporter),
                )

                if "_id" not in uploadObj:
                    raise Exception(
                        "After uploading a file chunk, did not receive object with _id. "
                        "Got instead: " + json.dumps(uploadObj)
                    )

                offset += len(chunk)

                if callable(progressCallback):
                    progressCallback({"current": offset, "total": size})

        if offset != size:
            self.delete("file/upload/" + uploadId)
            raise gcl.IncorrectUploadLengthError(
                "Expected upload to be %d bytes, but received %d." % (size, offset),
                upload=uploadObj,
            )

        return uploadObj


def _harmonize_girder_dandiset_to_dandi_api(rec):
    """
    Compare API (on a released version):

    {'count': 1,
     'created': '2020-07-21T22:22:15.396171Z',
     'dandiset': {'created': '2020-07-21T22:22:14.732729Z',
                  'identifier': '000027',
                  'updated': '2020-07-21T22:22:14.732762Z'},
     'metadata': {'dandiset': {...}},
     'updated': '2020-07-21T22:22:15.396295Z',
     'version': '0.200721.2222'}

    to Girder (on drafts):

    {'attrs': {'ctime': '2020-07-08T21:54:42.543000+00:00',
               'mtime': '2020-07-21T22:02:34.918000+00:00',
               'size': 0},
     'id': '5f0640a2ab90ac46c4561e4f',
     'metadata': {'dandiset': {...}},
     'name': '000027',
     'path': '000027',
     'type': 'dandiset'}

    So we will place some girder specific ones under 'girder' and populate 'dandiset', e.g.
    (there is absent clarify of what date times API returns:
    https://github.com/dandi/dandi-publish/issues/107
    so we will assume that my take was more or less correct and then we would have them
    correspond in case of a draft, as it is served by girder ATM:

    {# 'count': 1,  # no count
     'created': '2020-07-21T22:22:15.396171Z',  # attrs.ctime
     'dandiset': {'created': '2020-07-08T21:54:42.543000+00:00',  # attrs.ctime
                  'identifier': '000027',  # name
                  'updated': '2020-07-21T22:02:34.918000+00:00' },  # attrs.mtime
     'metadata': {'dandiset': {...}},
     'updated': '2020-07-21T22:02:34.918000+00:00'}  # attrs.mtime

    Parameters
    ----------
    rec

    Returns
    -------
    dict
    """
    # ATM it is just a simple remapping but might become more sophisticated later on
    return remap_dict(
        rec,
        {
            "metadata": "metadata.dandiset",  # 1-to-1 for now
            "dandiset.created": "attrs.ctime",
            "created": "attrs.ctime",
            "dandiset.uptimed": "attrs.mtime",
            "updated": "attrs.mtime",
            "dandiset.identifier": "name",
        },
    )


def _get_file_mtime(attrs):
    if not attrs:
        return None
    # We would rely on uploaded_mtime from metadata being stored as mtime.
    # If that one was not provided, the best we know is the "ctime"
    # for the file, use that one
    return ensure_datetime(attrs.get("mtime", attrs.get("ctime", None)))


def _harmonize_girder_asset_to_dandi_api(rec):
    """
    girder rec:

    *(Pdb) pprint(_a[0])
    {'attrs': {'ctime': '2020-07-21T22:00:36.362000+00:00',
               'mtime': '2020-07-21T17:31:55.283394-04:00',
               'size': 18792},
     'id': '5f176584f63d62e1dbd06946',
     'metadata': {... identical at this level
                  'uploaded_by': 'dandi 0.5.0+12.gd4ef762.dirty',
                  'uploaded_datetime': '2020-07-21T18:00:36.703727-04:00',
                  'uploaded_mtime': '2020-07-21T17:31:55.283394-04:00',
                  'uploaded_size': 18792},
     'name': 'sub-RAT123.nwb',
     'path': '000027/sub-RAT123/sub-RAT123.nwb',
     'type': 'file'}

    and API (lacking clear "modified" so needs tuning too):

        {
          "version": {
            "dandiset": {
              "identifier": "000027",
              "created": "2020-07-21T22:22:14.732729Z",
              "updated": "2020-07-21T22:22:14.732762Z"
            },
            "version": "0.200721.2222",
            "created": "2020-07-21T22:22:15.396171Z",
            "updated": "2020-07-21T22:22:15.396295Z",
            "count": 1
          },
          "uuid": "bca53c42-7fc2-41b6-b836-5ed102ba8447",
          "path": "/sub-RAT123/sub-RAT123.nwb",
          "size": 18792,
          "sha256": "1a765509384ea96b7b12136353d9c5b94f23d764ad0431e049197f7875eb352c",
          "created": "2020-07-21T22:22:16.882594Z",
          "updated": "2020-07-21T22:22:16.882641Z",
          "metadata": {
    ...
            "sha256": "1a765509384ea96b7b12136353d9c5b94f23d764ad0431e049197f7875eb352c",
    ...
            "uploaded_size": 18792,
            "uploaded_mtime": "2020-07-21T17:31:55.283394-04:00",
            "uploaded_datetime": "2020-07-21T18:00:36.703727-04:00",
    ...
          }
        }


    Parameters
    ----------
    rec

    Returns
    -------
    """
    rec = rec.copy()  # we will modify in place

    metadata = rec.get("metadata", {})
    size = rec["size"] = rec.get("attrs", {}).get("size")
    # we will add messages leading to decision that metadata is outdated and thus should not be used
    metadata_outdated = []
    uploaded_size = metadata.get("uploaded_size")
    if size is None:
        lgr.debug("Found no size in attrs from girder!")
        if uploaded_size is not None:
            lgr.debug("Taking 'uploaded_size' of %d", uploaded_size)
            rec["size"] = uploaded_size
    else:
        if uploaded_size is not None and size != uploaded_size:
            metadata_outdated.append(
                f"uploaded_size of {uploaded_size} != size of {size}"
            )

    # duplication but ok for now
    modified = rec["modified"] = _get_file_mtime(rec.get("attrs"))
    uploaded_mtime = metadata.get("uploaded_mtime")
    if uploaded_mtime:
        uploaded_mtime = ensure_datetime(uploaded_mtime)
    if modified is None:
        lgr.debug("Found no mtime (modified) among girder attrs")
        if uploaded_mtime is not None:
            rec["modified"] = uploaded_mtime
    else:
        if uploaded_mtime is not None and modified != uploaded_mtime:
            metadata_outdated.append(
                f"uploaded_mtime of {uploaded_mtime} != mtime of {modified}"
            )

    if metadata_outdated:
        lgr.warning(
            "Found discrepancies in girder record and metadata: %s",
            ", ".join(metadata_outdated),
        )

    # we need to strip off the leading dandiset identifier from the path
    path = rec["path"]
    if path.startswith("00"):
        # leading / is for consistency with API although yoh dislikes it
        # https://github.com/dandi/dandi-publish/issues/109
        # Girder client returned paths are OS specific.
        path = "/" + path.split(op.sep, 1)[1]
    else:
        lgr.debug(
            "Unexpected: an asset path did not have leading dandiset identifier: %s",
            path,
        )
    rec["path"] = path

    if "id" in rec:
        # Let's create a dedicated section for girder specific information
        rec["girder"] = {"id": rec.pop("id")}

    # Some additional fields which should appear at the top level in the records returned
    # by DANDI API
    if "sha256" in metadata:
        rec["sha256"] = metadata["sha256"]
    return rec


# TODO: our adapter on top of the Girder's client to simplify further
def get_client(server_url, authenticate=True, progressbars=False):
    """Simple authenticator which would store credential (api key) via keyring

    Parameters
    ----------
    server_url: str
      URL to girder instance
    """
    lgr.debug(f"Establishing a client for {server_url}")

    client = GirderCli(server_url, progressbars=progressbars)
    if authenticate:
        client.dandi_authenticate()
    return client


def get_collection(client, collection):
    return lookup(client, name=collection)


def ensure_collection(client, collection):
    try:
        return lookup(client, name=collection)
    except GirderNotFound:
        lgr.info(f"Collection {collection} was not found, creating")
        # TODO: split away or provide UI to tune (private, description)
        return client.createCollection(collection, public=True)


def ensure_folder(client, collection_rec, collection, folder):
    """Ensure presence of the folder.  If not present -- create all leading

    TODO:
     - ATM doesn't care about providing `description` and `public` options
     - doesn't check the type in folder_rec -- may be a file???
    """
    assert folder and str(folder) not in (
        "/",
        ".",
    ), "Folder must not be empty, and should not be . or /"
    try:
        folder_rec = lookup(client, name=collection, path=folder)
    except GirderNotFound:
        # We need to create all the leading paths if not there yet, and arrive
        # to the target folder.
        # We will rely on memoization over lookup for all the folder_rec querying
        parent_id = collection_rec["_id"]
        parent_type = "collection"
        parent_path = Path()
        for parent_dirname in folder.parts or ("",):
            parent_path /= parent_dirname
            try:
                folder_rec = lookup(client, name=collection, path=parent_path)
            except GirderNotFound:
                lgr.debug(f"Forder {parent_dirname} was not found, creating")
                folder_rec = client.createFolder(
                    parent_id,
                    parent_dirname,
                    parentType=parent_type,
                    # for now just depend on collection setup public=True,
                )
            parent_id = folder_rec["_id"]
            parent_type = "folder"
    lgr.debug(f"Folder: {folder_rec}")
    return folder_rec


class TQDMProgressReporter(object):
    reportProgress = True

    def __init__(self, label="", length=0):
        import tqdm

        self.label = label
        self.length = length

        self._pbar = None
        try:
            self._pbar = tqdm.tqdm(desc=label, total=length, unit="B", unit_scale=True)
        except AssertionError as exc:
            lgr.warning(
                "No progress indication for %s. Failed to initiate tqdm progress bar: %s",
                label,
                exc,
            )

    def update(self, chunkSize):
        if _DANDI_LOG_GIRDER:
            lgr.debug("PROGRESS[%s]: +%d", id(self), chunkSize)
        if self._pbar:
            self._pbar.update(chunkSize)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if self._pbar:
            self._pbar.clear()  # remove from screen -- not in effect ATM TODO
            self._pbar.close()
            del self._pbar


def keyring_lookup(service_name, username):
    """
    Determine a keyring backend to use for storing & retrieving credentials as
    follows:

    - If the user has specified a backend explicitly via the
      ``PYTHON_KEYRING_BACKEND`` environment variable or a ``keyringrc.cfg``
      file, use that backend without checking whether it's usable (If it's not,
      the user messed up).

    - Otherwise, query the default backend (which is guaranteed to already have
      the requisite dependencies installed) for the credentials for the given
      service name and username.  If this completes without error (regardless
      of whether the backend contains any such credentials), use that backend.

    - If the query fails (e.g., because a GUI is required but the session is in
      a plain terminal), try using the ``EncryptedKeyring`` backend.

      - If the default backend *was* the ``EncryptedKeyring`` backend, error.

      - If the ``EncryptedKeyring`` backend is not in the list of available
        backends (likely because its dependencies are not installed, though
        that shouldn't happen if dandi was installed properly), error.

      - If ``EncryptedKeyring``'s data file already exists, use it as the
        backend.

      - If ``EncryptedKeyring``'s data file does not already exist, ask the
        user whether they want to start using ``EncryptedKeyring``.  If yes,
        then set ``keyringrc.cfg`` (if it does not already exist) to specify it
        as the default backend, and return the backend.  If no, error.

    Returns a keyring backend and the password it holds (if any) for the given
    service and username.
    """

    kb = load_env() or load_config()
    if kb:
        return (kb, kb.get_password(service_name, username))
    kb = get_keyring()
    try:
        password = kb.get_password(service_name, username)
    except KeyringError as e:
        lgr.info("Default keyring errors on query: %s", e)
        if isinstance(kb, EncryptedKeyring):
            lgr.info(
                "Default keyring is EncryptedKeyring; abandoning keyring procedure"
            )
            raise
        # Use `type(..) is` instead of `isinstance()` to weed out subclasses
        kbs = [k for k in get_all_keyring() if type(k) is EncryptedKeyring]
        assert (
            len(kbs) == 1
        ), "EncryptedKeyring not available; is pycryptodomex installed?"
        kb = kbs[0]
        if op.exists(kb.file_path):
            lgr.info("EncryptedKeyring file exists; using as keyring backend")
            return (kb, kb.get_password(service_name, username))
        lgr.info("EncryptedKeyring file does not exist")
        if click.confirm(
            "Would you like to establish an encrypted keyring?", default=True
        ):
            keyring_cfg = Path(keyringrc_file())
            if keyring_cfg.exists():
                lgr.info("%s exists; refusing to overwrite", keyring_cfg)
            else:
                lgr.info(
                    "Configuring %s to use EncryptedKeyring as default backend",
                    keyring_cfg,
                )
                keyring_cfg.parent.mkdir(parents=True, exist_ok=True)
                keyring_cfg.write_text(
                    "[backend]\n"
                    "default-keyring = keyrings.alt.file.EncryptedKeyring\n"
                )
            return (kb, None)
        raise
    else:
        return (kb, password)


def keyringrc_file():
    return op.join(config_root(), "keyringrc.cfg")



import os
from pathlib import Path
import random
import re

import pytest
import responses

from ..consts import collection_drafts, dandiset_metadata_file
from ..dandiapi import DandiAPIClient
from ..dandiset import Dandiset
from ..download import download
from .. import girder
from ..register import register
from ..upload import upload
from ..utils import find_files, yaml_load


def test_upload(local_docker_compose_env, monkeypatch, organized_nwb_dir2):
    nwb_files = list(organized_nwb_dir2.glob(f"*{os.sep}*.nwb"))
    assert len(nwb_files) == 2
    dirname1 = nwb_files[0].parent.name
    filename1 = nwb_files[0].name
    dirname2 = nwb_files[1].parent.name
    assert dirname1 != dirname2

    dandi_instance_id = local_docker_compose_env["instance_id"]

    register(
        "Upload Test",
        "Upload Test Description",
        dandiset_path=organized_nwb_dir2,
        dandi_instance=dandi_instance_id,
    )
    with (organized_nwb_dir2 / dandiset_metadata_file).open() as fp:
        metadata = yaml_load(fp, typ="safe")
    dandi_id = metadata["identifier"]

    client = girder.get_client(local_docker_compose_env["instance"].girder)
    for dname in [dirname1, dirname2]:
        with pytest.raises(girder.GirderNotFound):
            girder.lookup(client, collection_drafts, path=f"{dandi_id}/{dname}")

    monkeypatch.chdir(organized_nwb_dir2)
    upload(paths=[dirname1], dandi_instance=dandi_instance_id, devel_debug=True)

    girder.lookup(client, collection_drafts, path=f"{dandi_id}/{dirname1}/{filename1}")
    with pytest.raises(girder.GirderNotFound):
        girder.lookup(client, collection_drafts, path=f"{dandi_id}/{dirname2}")


def test_upload_existing_error(
    local_docker_compose_env, monkeypatch, organized_nwb_dir
):
    dandi_instance_id = local_docker_compose_env["instance_id"]
    register(
        "Upload Test",
        "Upload Test Description",
        dandiset_path=organized_nwb_dir,
        dandi_instance=dandi_instance_id,
    )
    monkeypatch.chdir(organized_nwb_dir)
    upload(
        paths=[], dandi_instance=dandi_instance_id, devel_debug=True, existing="error"
    )
    with pytest.raises(FileExistsError):
        upload(
            paths=[],
            dandi_instance=dandi_instance_id,
            devel_debug=True,
            existing="error",
        )


def test_upload_locks(local_docker_compose_env, mocker, monkeypatch, organized_nwb_dir):
    (nwb_file,) = organized_nwb_dir.glob(f"*{os.sep}*.nwb")
    dirname = nwb_file.parent.name
    dandi_instance_id = local_docker_compose_env["instance_id"]
    register(
        "Upload Test",
        "Upload Test Description",
        dandiset_path=organized_nwb_dir,
        dandi_instance=dandi_instance_id,
    )
    monkeypatch.chdir(organized_nwb_dir)
    lockmock = mocker.patch.object(girder.GirderCli, "lock_dandiset")
    upload(
        paths=[dirname],
        dandi_instance=dandi_instance_id,
        devel_debug=True,
        existing="error",
    )
    lockmock.assert_called()
    lockmock.return_value.__enter__.assert_called()
    lockmock.return_value.__exit__.assert_called()


def test_upload_unregistered(local_docker_compose_env, monkeypatch, organized_nwb_dir):
    (nwb_file,) = organized_nwb_dir.glob(f"*{os.sep}*.nwb")
    dirname = nwb_file.parent.name
    dandi_instance_id = local_docker_compose_env["instance_id"]
    (organized_nwb_dir / dandiset_metadata_file).write_text("identifier: '999999'\n")
    monkeypatch.chdir(organized_nwb_dir)
    with pytest.raises(ValueError) as excinfo:
        upload(paths=[dirname], dandi_instance=dandi_instance_id, devel_debug=True)
    assert str(excinfo.value) == (
        f"There is no 999999 in {collection_drafts}. Did you use 'dandi register'?"
    )


def test_upload_path_not_in_dandiset_path(
    local_docker_compose_env, organized_nwb_dir, organized_nwb_dir2
):
    (organized_nwb_dir / dandiset_metadata_file).write_text("identifier: '000001'\n")
    with pytest.raises(ValueError) as excinfo:
        upload(
            paths=[str(organized_nwb_dir2)],
            dandiset_path=organized_nwb_dir,
            dandi_instance=local_docker_compose_env["instance_id"],
        )
    assert re.fullmatch(
        rf"{re.escape(str(organized_nwb_dir2))}\S* is not under "
        rf"{re.escape(str(organized_nwb_dir))}",
        str(excinfo.value),
    )


def test_upload_nonexistent_path(local_docker_compose_env, organized_nwb_dir):
    # This is currently just a smoke test to cover the "except
    # FileNotFoundError:" block in process_path().
    (organized_nwb_dir / dandiset_metadata_file).write_text("identifier: '000001'\n")
    upload(
        paths=[str(organized_nwb_dir / "nonexistent.nwb")],
        dandiset_path=organized_nwb_dir,
        dandi_instance=local_docker_compose_env["instance_id"],
    )


def test_upload_stat_failure(local_docker_compose_env, organized_nwb_dir):
    # This is currently just a smoke test to cover the "except Exception:"
    # block in process_path().
    (organized_nwb_dir / dandiset_metadata_file).write_text("identifier: '000001'\n")
    baddir = organized_nwb_dir / "bad"
    baddir.mkdir(mode=0o444)
    try:
        upload(
            paths=[str(baddir / "nonexistent.nwb")],
            dandiset_path=organized_nwb_dir,
            dandi_instance=local_docker_compose_env["instance_id"],
        )
    finally:
        baddir.rmdir()


def test_upload_external_download(local_docker_compose_env, monkeypatch, tmp_path):
    download("https://dandiarchive.org/dandiset/000027", tmp_path)
    dandiset_path = tmp_path / "000027"
    dandi_instance_id = local_docker_compose_env["instance_id"]
    dandiset = Dandiset(dandiset_path)

    # Since identifier is instance-specific and locking requires dandiset to exist
    # we cannot just upload.  We need first to register a new one and have identifier updated.
    # To minimize any programmatic changes, we will not point register to the dandiset
    # but just will use its  "identifier"  in the simplest .replace
    rec = register(
        dandiset.metadata["name"],
        dandiset.metadata["description"],
        dandi_instance=dandi_instance_id,
    )
    dandiset_yaml = dandiset_path / dandiset_metadata_file
    new_dandiset_yaml = dandiset_yaml.read_text().replace("000027", rec["identifier"])
    dandiset_yaml.write_text(new_dandiset_yaml)

    monkeypatch.chdir(dandiset_path)
    upload(paths=[], dandi_instance=dandi_instance_id, devel_debug=True)


def test_upload_size_mismatch(
    capsys, local_docker_compose_env, monkeypatch, organized_nwb_dir
):
    (nwb_file,) = organized_nwb_dir.glob(f"*{os.sep}*.nwb")
    dirname = nwb_file.parent.name
    dandi_instance_id = local_docker_compose_env["instance_id"]
    register(
        "Upload Test",
        "Upload Test Description",
        dandiset_path=organized_nwb_dir,
        dandi_instance=dandi_instance_id,
    )
    monkeypatch.chdir(organized_nwb_dir)
    with responses.RequestsMock() as rsps:
        rsps.add_passthru(re.compile(r".+"))
        rsps.add(
            responses.GET,
            re.compile(
                r"^{}/api/v1/file/[^/]+/download$".format(
                    re.escape(local_docker_compose_env["instance"].girder)
                )
            ),
            body="Disregard this.",
            headers={"Content-Length": "42"},
            stream=True,
        )
        upload(paths=[dirname], dandi_instance=dandi_instance_id, devel_debug=True)
    assert (
        repr(
            {
                "status": "skipped",
                "message": "File size on server does not match local file",
            }
        )
        in capsys.readouterr().out
    )


@pytest.mark.skipif(
    not os.environ.get("DANDI_DEVEL"), reason="Only run when DANDI_DEVEL is set"
)
def test_enormous_upload_breaks_girder(
    capsys, local_docker_compose_env, monkeypatch, tmp_path
):
    dandi_instance_id = local_docker_compose_env["instance_id"]
    register(
        "Enormous File Upload Test",
        "Enormous File Upload Test Description",
        dandiset_path=tmp_path,
        dandi_instance=dandi_instance_id,
    )
    bigfile = tmp_path / "blob.dat"
    meg = bytes(random.choices(range(256), k=1 << 20))
    try:
        with bigfile.open("wb") as fp:
            for _ in range(66 * 1024):
                fp.write(meg)
        monkeypatch.chdir(tmp_path)
        upload(
            paths=[str(bigfile)],
            dandi_instance=dandi_instance_id,
            allow_any_path=True,
            devel_debug=True,
        )
        assert "'status': 'skipped'" in capsys.readouterr().out
    finally:
        try:
            bigfile.unlink()
        except FileNotFoundError:
            pass


def test_new_upload_download(local_dandi_api, monkeypatch, organized_nwb_dir, tmp_path):
    r = local_dandi_api["client"].create_dandiset("Test Dandiset", {})
    dandiset_id = r["identifier"]
    (nwb_file,) = organized_nwb_dir.glob(f"*{os.sep}*.nwb")
    (organized_nwb_dir / dandiset_metadata_file).write_text(
        f"identifier: '{dandiset_id}'\n"
    )
    monkeypatch.chdir(organized_nwb_dir)
    monkeypatch.setenv("DANDI_API_KEY", local_dandi_api["api_key"])
    upload(paths=[], dandi_instance=local_dandi_api["instance_id"], devel_debug=True)
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/draft",
        tmp_path,
    )
    (nwb_file2,) = tmp_path.glob(f"{dandiset_id}{os.sep}*{os.sep}*.nwb")
    assert nwb_file.name == nwb_file2.name
    assert nwb_file.parent.name == nwb_file2.parent.name

    #
    # test updating dandiset metadata record while at it
    # For now let's "manually" populate dandiset.yaml in that downloaded location
    # which is missing due to https://github.com/dandi/dandi-api/issues/63
    from ..dandiset import APIDandiset
    from ..utils import yaml_dump

    ds_orig = APIDandiset(organized_nwb_dir)
    ds_metadata = ds_orig.metadata
    ds_metadata["description"] = "very long"
    ds_metadata["name"] = "shorty"

    monkeypatch.chdir(tmp_path / dandiset_id)
    Path(dandiset_metadata_file).write_text(yaml_dump(ds_metadata))
    upload(
        paths=[dandiset_metadata_file],
        dandi_instance=local_dandi_api["instance_id"],
        devel_debug=True,
        upload_dandiset_metadata=True,
    )

    r = local_dandi_api["client"].get_dandiset(dandiset_id, "draft")
    assert r["metadata"]["name"] == "shorty"


def test_new_upload_extant_existing(mocker, text_dandiset):
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    with pytest.raises(FileExistsError):
        text_dandiset["reupload"](existing="error")
    iter_upload_spy.assert_not_called()


def test_new_upload_extant_skip(mocker, text_dandiset):
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing="skip")
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize("existing", ["overwrite", "refresh"])
def test_new_upload_extant_eq_overwrite(existing, mocker, text_dandiset):
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing=existing)
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize("existing", ["overwrite", "refresh"])
def test_new_upload_extant_neq_overwrite(
    existing, local_dandi_api, mocker, text_dandiset, tmp_path
):
    dandiset_id = text_dandiset["dandiset_id"]
    (text_dandiset["dspath"] / "file.txt").write_text("This is different text.\n")
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing=existing)
    iter_upload_spy.assert_called()
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/draft",
        tmp_path,
    )
    assert (
        tmp_path / dandiset_id / "file.txt"
    ).read_text() == "This is different text.\n"


def test_new_upload_extant_old_refresh(mocker, text_dandiset):
    (text_dandiset["dspath"] / "file.txt").write_text("This is different text.\n")
    os.utime(text_dandiset["dspath"] / "file.txt", times=(0, 0))
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing="refresh")
    iter_upload_spy.assert_not_called()


def test_new_upload_extant_force(mocker, text_dandiset):
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing="force")
    iter_upload_spy.assert_called()


def test_new_upload_extant_bad_existing(mocker, text_dandiset):
    iter_upload_spy = mocker.spy(DandiAPIClient, "iter_upload")
    text_dandiset["reupload"](existing="foobar")
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize(
    "contents",
    [
        pytest.param(
            b"",
            marks=pytest.mark.xfail(
                reason="https://github.com/dandi/dandi-api/issues/168"
            ),
        ),
        b"x",
    ],
)
def test_upload_download_small_file(contents, local_dandi_api, monkeypatch, tmp_path):
    client = local_dandi_api["client"]
    dandiset_id = client.create_dandiset("Small Dandiset", {})["identifier"]
    dspath = tmp_path / "upload"
    dspath.mkdir()
    (dspath / dandiset_metadata_file).write_text(f"identifier: '{dandiset_id}'\n")
    (dspath / "file.txt").write_bytes(contents)
    monkeypatch.setenv("DANDI_API_KEY", local_dandi_api["api_key"])
    upload(
        paths=[],
        dandiset_path=dspath,
        dandi_instance=local_dandi_api["instance_id"],
        devel_debug=True,
        allow_any_path=True,
        validation="skip",
    )
    download_dir = tmp_path / "download"
    download_dir.mkdir()
    download(
        f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}/versions/draft",
        download_dir,
    )
    files = sorted(map(Path, find_files(r".*", paths=[download_dir])))
    assert files == [
        download_dir / dandiset_id / dandiset_metadata_file,
        download_dir / dandiset_id / "file.txt",
    ]
    assert files[1].read_bytes() == contents



"""
Interactions with DANDI archive ATM go either through Girder or through DANDI API.

This module provides helpers to decide on which API Client to use and/or
abstract higher level code from specifics of any particular client.

Eventually it is largely to be "dissolved" whenever we stop talking to girder.
"""

from contextlib import contextmanager
import re
from urllib.parse import unquote as urlunquote

import requests

from .consts import known_instances, known_instances_rev
from .dandiapi import DandiAPIClient
from .exceptions import FailedToConnectError, NotFoundError, UnknownURLError
from . import get_logger, girder
from .utils import get_instance

lgr = get_logger()


@contextmanager
def navigate_url(url):
    """Context manager to 'navigate' URL pointing to DANDI archive.

    Parameters
    ----------
    url: str
      URL which might point to a dandiset, a folder, or an asset(s)

    Yields
    ------
    client, dandiset, assets (generator)
      `client` will have established a session for the duration of the context
    """
    server_type, server_url, asset_type, asset_id = parse_dandi_url(url)

    # We could later try to "dandi_authenticate" if run into permission issues.
    # May be it could be not just boolean but the "id" to be used?
    kwargs = {}
    if server_type == "girder":
        asset_id, asset_type, client, server_type = _map_to_girder(url)
        args = asset_id, asset_type
    elif server_type == "api":
        client = DandiAPIClient(server_url)
        if asset_id["version"] is None:
            r = client.get(f"/dandisets/{asset_id['dandiset_id']}/")
            if "draft_version" in r:
                asset_id["version"] = r["draft_version"]["version"]
                published_version = r["most_recent_published_version"]
                if published_version:
                    asset_id["version"] = published_version["version"]
            else:
                # TODO: remove `if` after https://github.com/dandi/dandi-api/pull/219
                # is merged/deployed
                asset_id["version"] = r["most_recent_version"]["version"]
        args = (asset_id["dandiset_id"], asset_id["version"])
        kwargs["include_metadata"] = True
        if asset_id.get("location") or asset_id.get("asset_id"):
            with client.session():
                dandiset = client.get_dandiset(*args)
                if asset_type == "folder":
                    assets = client.get_dandiset_assets(
                        *args, path=asset_id["location"]
                    )
                elif asset_type == "item":
                    if "location" in asset_id:
                        asset = client.get_asset_bypath(*args, asset_id["location"])
                        assets = [asset] if asset is not None else []
                    else:
                        assets = [client.get_asset(*args, asset_id["asset_id"])]
                else:
                    raise NotImplementedError(
                        f"Do not know how to handle asset type {asset_type} with location"
                    )
                yield (client, dandiset, assets)
            return
    else:
        raise NotImplementedError(
            f"Download from server of type {server_type} is not yet implemented"
        )

    with client.session():
        dandiset, assets = client.get_dandiset_and_assets(*args, **kwargs)
        yield client, dandiset, assets


def _map_to_girder(url):
    """
    discover girder_id for a draft dataset
    """
    # This is a duplicate call from above but it is cheap, so decided to just redo
    # it here instead of passing all the variables + url
    server_type, server_url, asset_type, asset_id = parse_dandi_url(url)
    server_url = known_instances[known_instances_rev[server_url.rstrip("/")]].girder
    client = girder.get_client(server_url, authenticate=False, progressbars=True)
    # TODO: RF if https://github.com/dandi/dandiarchive/issues/316 gets addressed
    # A hybrid UI case not yet adjusted for drafts API.
    # TODO: remove whenever it is gone in an unknown version
    if asset_id.get("folder_id"):
        asset_type = "folder"
        asset_id = [asset_id.get("folder_id")]
    else:
        girder_path = "{}/{}".format("drafts", asset_id["dandiset_id"])
        asset_type = "folder"
        if asset_id.get("location"):
            # Not implemented by UI ATM but might come
            girder_path = "{}/{}".format(girder_path, asset_id["location"])
        try:
            girder_rec = girder.lookup(client, girder_path)
        except BaseException:
            lgr.warning(f"Failed to lookup girder information for {girder_path}")
            girder_rec = None
        if not girder_rec:
            raise RuntimeError(f"Cannot download from {url}")
        asset_id = girder_rec.get("_id")
    return asset_id, asset_type, client, "girder"


class _dandi_url_parser:
    # Defining as a class with all the attributes to not leak all the variables
    # etc into module space, and later we might end up with classes for those
    # anyways
    id_regex = "[a-f0-9]{24}"
    id_grp = f"(?P<id>{id_regex})"
    dandiset_id_grp = "(?P<dandiset_id>[0-9]{6})"
    # Should absorb port and "api/":
    server_grp = "(?P<server>(?P<protocol>https?)://(?P<hostname>[^/]+)/(api/)?)"
    known_urls = [
        # List of (regex, settings, display string) triples
        #
        # Settings:
        # - handle_redirect:
        #   - 'pass' - would continue with original url if no redirect happen
        #   - 'only' - would interrupt if no redirection happens
        # - server_type:
        #   - 'girder' - underlying requests should go to girder server
        #   - 'api' - the "new" API service
        # - rewrite:
        #   - callable -- which would rewrite that "URI"
        # - map_instance
        #
        # Those we first redirect and then handle the redirected URL
        # TODO: Later should better conform to our API, so we could allow
        #       for not only "dandiarchive.org" URLs
        (
            re.compile(r"DANDI:.*"),
            {"rewrite": lambda x: "https://identifiers.org/" + x},
            "DANDI:<dandiset id>",
        ),
        (
            re.compile(r"https?://dandiarchive\.org/.*"),
            {"handle_redirect": "pass"},
            "https://dandiarchive.org/...",
        ),
        (
            re.compile(r"https?://identifiers\.org/DANDI:.*"),
            {"handle_redirect": "pass"},
            "https://identifiers.org/DANDI:<dandiset id>",
        ),
        # New DANDI API, ATM can be reached only via enable('DJANGO_API')  in
        # browser console
        # https://gui.dandiarchive.org/#/dandiset/000001/0.201104.2302/files
        # TODO: upload something to any dandiset to see what happens when there
        # are files and adjust for how path is provided (if not ?location=)
        (
            re.compile(
                r"(?P<server>(?P<protocol>https?)://"
                r"(?P<hostname>gui-beta-dandiarchive-org\.netlify\.app)/)"
                rf"#/(?P<asset_type>dandiset)/{dandiset_id_grp}"
                r"(/(?P<version>[.0-9]{5,}|draft))?"
                rf"(/files(\?location=(?P<location>.*)?)?)?"
            ),
            {"server_type": "api"},
            "https://gui-beta-dandiarchive-org.netflif.app/#/dandiset"
            "/<dandiset id>[/<version>][/files[?location=<path>]]",
        ),
        # PRs are also on netlify - so above takes precedence. TODO: make more
        # specific?
        (
            re.compile(r"https?://[^/]*dandiarchive-org\.netlify\.app/.*"),
            {"map_instance": "dandi"},
            "https://*dandiarchive-org.netflify.app/...",
        ),
        # Direct urls to our new API
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
                r"(/(versions(/(?P<version>[.0-9]{5,}|draft))?)?)?"
            ),
            {"server_type": "api"},
            "https://<server>[/api]/dandisets/<dandiset id>[/versions[/<version>]]",
        ),
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
                r"/versions/(?P<version>[.0-9]{5,}|draft)"
                r"/assets/(?P<asset_id>[^/]+)(/(download/?)?)?"
            ),
            {"server_type": "api"},
            "https://<server>[/api]/dandisets/<dandiset id>/versions/<version>"
            "/assets/<asset id>[/download]",
        ),
        # But for drafts files navigator it is a bit different beast and there
        # could be no versions, only draft
        # https://deploy-preview-341--gui-dandiarchive-org.netlify.app/#/dandiset/000027/draft/files?_id=5f176583f63d62e1dbd06943&_modelType=folder
        (
            re.compile(
                rf"{server_grp}#/(?P<asset_type>dandiset)/{dandiset_id_grp}"
                r"(/(?P<version>draft))?"
                rf"(/files(\?_id={id_grp}(&_modelType=folder)?)?)?"
            ),
            {"server_type": "girder"},
            "https://<server>[/api]#/dandiset/<dandiset id>[/draft]"
            "[/files[?_id=<id>[&_modelType=folder]]]",
        ),
        # ad-hoc explicitly pointing within URL to the specific instance to use
        # and otherwise really simple:
        # dandi://INSTANCE/DANDISET_ID[@VERSION][/PATH]
        # For now to not be promoted to the users, and primarily for internal
        # use
        (
            re.compile(
                rf"dandi://(?P<instance_name>({'|'.join(known_instances)}))"
                rf"/{dandiset_id_grp}"
                r"(@(?P<version>[.0-9]{5,}|draft))?"
                rf"(/(?P<location>.*)?)?"
            ),
            {"server_type": "api"},
            "dandi://<instance name>/<dandiset id>[@<version>][/<path>]",
        ),
        # https://deploy-preview-341--gui-dandiarchive-org.netlify.app/#/dandiset/000006/draft
        # (no API yet)
        (
            re.compile(r"https?://.*"),
            {"handle_redirect": "only"},
            "https://<server>/...",
        ),
    ]
    # We might need to remap some assert_types
    map_asset_types = {"dandiset": "folder"}
    # And lets create our mapping into girder instances from known_instances:
    map_to = {"girder": {}, "api": {}}
    for (
        metadata_version,
        girder,  # noqa: F402
        gui,
        redirector,
        api,
    ) in known_instances.values():
        for h in (gui, redirector):
            if h:
                if girder:
                    map_to["girder"][h] = girder
                if api:
                    map_to["api"][h] = api

    @classmethod
    def parse(cls, url, *, map_instance=True):
        """Parse url like and return server (address), asset_id and/or directory

        Example URLs (as of 20200310):
        - User public: (Users -> bendichter/Public/Tolias2020)
          [seems to be visible only if logged in]
          old (girder inflicted): https://gui.dandiarchive.org/#/folder/5e5593cc1a343161ff7c5a92
        - Collection top level (Collections -> yarik):
          https://gui.dandiarchive.org/#/collection/5daa5ca7e3489855a3027682
        - Collections: (Collections -> yarik/svoboda)
          old (girder inflicted): https://gui.dandiarchive.org/#/folder/5dab0830f377535c7d96c2b4
        - Dataset landing page metadata
          old (girder inflicted): https://gui.dandiarchive.org/#/dandiset/5e6d5c6976569eb93f451e4f
          now (20210119): https://gui.dandiarchive.org/#/dandiset/000003

        Individual and multiple files:
          - dandi???

        Multiple selected files + folders -- we do not support ATM, then further
        RFing would be due, probably making this into a generator or returning a
        list of entries.

        "Features":

        - uses some of `known_instance`s to map some urls, e.g. from
          gui.dandiarchive.org ones into girder.

        Returns
        -------
        server_type, server, asset_type, asset_id
          asset_type is either asset_id or folder ATM. asset_id might be a list
          in case of multiple files

        """
        lgr.debug("Parsing url %s", url)

        # Loop through known url regexes and stop as soon as one is matching
        match = None
        for regex, settings, _ in cls.known_urls:
            match = regex.fullmatch(url)
            if not match:
                continue
            groups = match.groupdict()
            lgr.log(5, "Matched %r into %s", url, groups)
            rewrite = settings.get("rewrite", False)
            handle_redirect = settings.get("handle_redirect", False)
            if rewrite:
                assert not handle_redirect
                assert not settings.get("map_instance")
                new_url = rewrite(url)
                return cls.parse(new_url)
            elif handle_redirect:
                assert handle_redirect in ("pass", "only")
                new_url = cls.follow_redirect(url)
                if new_url != url:
                    return cls.parse(new_url)
                if handle_redirect == "pass":
                    # We used to issue warning in such cases, but may be it got implemented
                    # now via reverse proxy and we had added a new regex? let's just
                    # continue with a debug msg
                    lgr.debug("Redirection did not happen for %s", url)
                else:
                    raise RuntimeError(
                        f"{url} did not redirect to another location which dandi client would"
                        f" know how to handle."
                    )
            elif settings.get("map_instance"):
                if map_instance:
                    server_type, server, *_ = cls.parse(url, map_instance=False)
                    if settings["map_instance"] not in known_instances:
                        raise ValueError(
                            "Unknown instance {}. Known are: {}".format(
                                settings["map_instance"], ", ".join(known_instances)
                            )
                        )
                    known_instance = get_instance(settings["map_instance"])
                    # for consistency, add
                    server = getattr(known_instance, server_type)
                    if not server.endswith("/"):
                        server += "/"
                    return (server_type, server) + tuple(_)
                continue  # in this run we ignore an match further
            elif "instance_name" in groups:
                known_instance = get_instance(groups["instance_name"])
                server_type = "girder" if known_instance.girder else "api"
                assert known_instance.api  # must be defined
                groups["server"] = known_instance.api
                # could be overloaded later depending if location is provided
                groups["asset_type"] = "dandiset"
                break
            else:
                server_type = settings.get("server_type", "girder")
                break
        if not match:
            known_regexes = "\n - ".join(
                [""] + [display for _, _, display in cls.known_urls]
            )
            # TODO: may be make use of etelemetry and report if newer client
            # which might know is available?
            raise UnknownURLError(
                f"We do not know how to map URL {url} to our servers.\n"
                f"Patterns for known setups:"
                f"{known_regexes}"
            )

        url_server = groups["server"]
        server = cls.map_to[server_type].get(url_server.rstrip("/"), url_server)

        if not server.endswith("/"):
            server += "/"  # we expected '/' to be there so let it be

        asset_type = groups.get("asset_type")
        dandiset_id = groups.get("dandiset_id")
        version = groups.get("version")
        location = groups.get("location")
        asset_key = groups.get("asset_id")
        if location:
            location = urlunquote(location)
            # ATM carries leading '/' which IMHO is not needed/misguiding somewhat, so
            # I will just strip it
            location = location.lstrip("/")
        if not (asset_type == "dandiset" and dandiset_id):
            raise ValueError(f"{url} does not point to a dandiset")
        if server_type == "girder" and not version:
            version = "draft"
        # Let's just return a structured record for the requested asset
        asset_ids = {"dandiset_id": dandiset_id, "version": version}
        # if location is not degenerate -- it would be a folder or a file
        if location:
            if location.endswith("/"):
                asset_type = "folder"
            else:
                asset_type = "item"
            asset_ids["location"] = location
        elif asset_key:
            asset_type = "item"
            asset_ids["asset_id"] = asset_key
        # TODO: remove whenever API supports "draft" and this type of url
        if groups.get("id"):
            assert version == "draft"
            asset_ids["folder_id"] = groups["id"]
            asset_type = "folder"

        ret = server_type, server, asset_type, asset_ids
        lgr.debug("Parsed into %s", ret)
        return ret

    @staticmethod
    def follow_redirect(url):
        r = requests.head(url, allow_redirects=True)
        if r.status_code == 404:
            raise NotFoundError(url)
        elif r.status_code != 200:
            raise FailedToConnectError(
                f"Response for getting {url} to redirect returned {r.status_code}."
                f" Please verify that it is a URL related to dandiarchive and"
                f" supported by dandi client"
            )
        elif r.url != url:
            return r.url
        return url


# convenience binding
parse_dandi_url = _dandi_url_parser.parse
follow_redirect = _dandi_url_parser.follow_redirect



from dataclasses import dataclass, field
from operator import attrgetter
from pathlib import Path, PurePosixPath
from typing import Iterable, Iterator, NamedTuple, Optional, Set, Tuple

import click
import requests

from .consts import known_instances
from .dandiapi import DandiAPIClient
from .dandiarchive import parse_dandi_url
from .exceptions import NotFoundError
from .utils import is_url


class RemoteAsset(NamedTuple):
    dandiset_id: str
    version_id: str
    asset_id: str
    path: str


@dataclass
class Deleter:
    """
    Class for registering assets & Dandisets to delete and then deleting them
    """

    client: Optional[DandiAPIClient] = None
    dandiset_id: Optional[str] = None
    #: Whether we are deleting an entire Dandiset (true) or just assets (false)
    deleting_dandiset: bool = False
    remote_assets: Set[RemoteAsset] = field(default_factory=set)

    def __bool__(self) -> bool:
        return self.deleting_dandiset or bool(self.remote_assets)

    def set_dandiset(self, api_url: str, dandiset_id: str) -> None:
        if self.client is None:
            self.client = DandiAPIClient(api_url)
            self.client.dandi_authenticate()
            self.dandiset_id = dandiset_id
            try:
                self.client.get_dandiset(dandiset_id, "draft")
            except requests.HTTPError as e:
                if e.response.status_code == 404:
                    raise NotFoundError(f"Dandiset {dandiset_id} not found on server")
                else:
                    raise
        elif not is_same_url(self.client.api_url, api_url):
            raise ValueError("Cannot delete assets from multiple API instances at once")
        elif self.dandiset_id != dandiset_id:
            raise ValueError("Cannot delete assets from multiple Dandisets at once")

    def register_dandiset(self, api_url: str, dandiset_id: str) -> None:
        self.set_dandiset(api_url, dandiset_id)
        self.deleting_dandiset = True

    def register_asset(
        self, api_url: str, dandiset_id: str, version_id: str, asset_path: str
    ) -> None:
        self.set_dandiset(api_url, dandiset_id)
        asset = self.client.get_asset_bypath(dandiset_id, version_id, asset_path)
        if asset is None:
            raise NotFoundError(
                f"Asset at path {asset_path!r} not found in Dandiset {dandiset_id}"
            )
        self.remote_assets.add(
            RemoteAsset(dandiset_id, version_id, asset["asset_id"], asset["path"])
        )

    def register_asset_folder(
        self, api_url: str, dandiset_id: str, version_id: str, folder_path: str
    ) -> None:
        self.set_dandiset(api_url, dandiset_id)
        any_assets = False
        for asset in self.client.get_dandiset_assets(
            dandiset_id, version_id, path=folder_path
        ):
            self.remote_assets.add(
                RemoteAsset(dandiset_id, version_id, asset["asset_id"], asset["path"])
            )
            any_assets = True
        if not any_assets:
            raise NotFoundError(
                f"No assets under path {folder_path!r} found in Dandiset {dandiset_id}"
            )

    def register_url(self, url: str) -> None:
        server_type, server_url, asset_type, asset_id = parse_dandi_url(url)
        if asset_type == "dandiset":
            if asset_id.get("version") is not None:
                raise NotImplementedError(
                    "Dandi API server does not support deletion of individual"
                    " versions of a dandiset"
                )
            self.register_dandiset(server_url, asset_id["dandiset_id"])
        elif asset_type == "item":
            self.register_asset(
                server_url,
                asset_id["dandiset_id"],
                asset_id.get("version") or "draft",
                asset_id["location"],
            )
        elif asset_type == "folder":
            self.register_asset_folder(
                server_url,
                asset_id["dandiset_id"],
                asset_id.get("version") or "draft",
                asset_id["location"],
            )
        else:
            raise RuntimeError(f"Unexpected asset type for {url}: {asset_type}")

    def register_local_path_equivalent(self, instance_name: str, filepath: str) -> None:
        instance = known_instances[instance_name]
        if instance.girder is not None:
            raise NotImplementedError("Cannot delete assets from Girder instances")
        api_url = instance.api
        dandiset_id, asset_path = find_local_asset(filepath)
        self.set_dandiset(api_url, dandiset_id)
        if asset_path.endswith("/"):
            self.register_asset_folder(api_url, dandiset_id, "draft", asset_path)
        else:
            self.register_asset(api_url, dandiset_id, "draft", asset_path)

    def confirm(self) -> bool:
        if self.deleting_dandiset:
            msg = f"Delete Dandiset {self.dandiset_id}?"
        else:
            msg = (
                f"Delete {len(self.remote_assets)} assets on server from"
                f" Dandiset {self.dandiset_id}?"
            )
        return click.confirm(msg)

    def delete_dandiset(self) -> None:
        if self.deleting_dandiset:
            self.client.delete_dandiset(self.dandiset_id)
        else:
            raise RuntimeError(
                "delete_dandiset() called when Dandiset not registered for deletion"
            )

    def _process_asset(self, asset: RemoteAsset) -> Iterator[dict]:
        yield {"status": "Deleting"}
        try:
            self.client.delete_asset(
                asset.dandiset_id, asset.version_id, asset.asset_id
            )
        except Exception as e:
            yield {"status": "Error", "message": f"{type(e).__name__}: {e}"}
        else:
            yield {"status": "Deleted"}

    def process_assets_pyout(self) -> Iterator[dict]:
        with self.client.session():
            for asset in sorted(self.remote_assets, key=attrgetter("path")):
                yield {
                    "path": asset.path,
                    ("status", "message"): self._process_asset(asset),
                }

    def process_assets_debug(self) -> Iterator[Iterator[dict]]:
        with self.client.session():
            for asset in sorted(self.remote_assets, key=attrgetter("path")):
                yield ({"path": asset.path, **d} for d in self._process_asset(asset))


def delete(
    paths: Iterable[str],
    dandi_instance: str = "dandi",
    devel_debug: bool = False,
    jobs: Optional[int] = None,
    force: bool = False,
) -> None:
    deleter = Deleter()
    for p in paths:
        if is_url(p):
            deleter.register_url(p)
        else:
            deleter.register_local_path_equivalent(dandi_instance, p)
    if deleter and (force or deleter.confirm()):
        if deleter.deleting_dandiset:
            deleter.delete_dandiset()
        elif devel_debug:
            for gen in deleter.process_assets_debug():
                for r in gen:
                    print(r, flush=True)
        else:
            from .support import pyout as pyouts

            pyout_style = pyouts.get_style(hide_if_missing=False)
            rec_fields = ("path", "status", "message")
            out = pyouts.LogSafeTabular(
                style=pyout_style, columns=rec_fields, max_workers=jobs
            )
            with out:
                for gen in deleter.process_assets_pyout():
                    out(gen)


def find_local_asset(filepath: str) -> Tuple[str, str]:
    """
    Given a path to a local file, return the ID of the Dandiset in which it is
    located and the path to the file relative to the root of said Dandiset.  If
    the file is a directory, the path will end with a trailing slash.
    """
    from .dandiset import Dandiset

    path = Path(filepath).absolute()
    dandiset = Dandiset.find(path.parent)
    relpath = str(PurePosixPath(path.relative_to(dandiset.path)))
    if path.is_dir():
        relpath += "/"
    return (dandiset.identifier, relpath)


def is_same_url(url1: str, url2: str) -> bool:
    # TODO: Use a real URL library like furl, purl, or yarl
    return url1.rstrip("/") == url2.rstrip("/")



import pytest

from dandi.consts import known_instances
from dandi.dandiarchive import follow_redirect, parse_dandi_url
from dandi.exceptions import NotFoundError
from dandi.tests.skip import mark


@pytest.mark.parametrize(
    "url,asset_type,asset_id",
    [
        # ATM we point to drafts, so girder
        ("DANDI:000027", "dandiset", {"dandiset_id": "000027", "version": "draft"}),
        # Example of current web ui (with girder backend) as of 20210119
        (
            "https://gui.dandiarchive.org/#/dandiset/000003/draft/"
            "files?_id=5e74ee41368d3c79a8006d29&_modelType=folder",
            "folder",
            {
                "dandiset_id": "000003",
                "version": "draft",
                "folder_id": "5e74ee41368d3c79a8006d29",
            },
        ),
        pytest.param(
            "https://dandiarchive.org/dandiset/000003",
            "dandiset",
            {"dandiset_id": "000003", "version": "draft"},
            marks=mark.skipif_no_network,
        ),
    ],
)
def test_parse_girder_url(url, asset_type, asset_id):
    st, s, a, aid = parse_dandi_url(url)
    assert st == "girder"
    assert s == known_instances["dandi"].girder + "/"
    assert a == asset_type
    assert aid == asset_id


@pytest.mark.parametrize(
    "url,instance,asset_type,asset_id",
    [
        # New DANDI web UI driven by DANDI API.
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/000001",
            "dandi-api",
            "dandiset",
            {"dandiset_id": "000001", "version": None},
        ),
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/000001/0.201104.2302",
            "dandi-api",
            "dandiset",
            {"dandiset_id": "000001", "version": "0.201104.2302"},
        ),
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/000001/0.201104.2302/files",
            "dandi-api",
            "dandiset",
            {"dandiset_id": "000001", "version": "0.201104.2302"},
        ),
        (
            "http://localhost:8000/api/dandisets/000002/versions/draft",
            "dandi-api-local-docker-tests",
            "dandiset",
            {"dandiset_id": "000002", "version": "draft"},
        ),
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/000001/"
            "files?location=%2Fsub-anm369962",
            "dandi-api",
            "item",
            {"dandiset_id": "000001", "version": None, "location": "sub-anm369962"},
        ),
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/000006/"
            "0.200714.1807/files?location=%2Fsub-anm369962",
            "dandi-api",
            "item",
            {
                "dandiset_id": "000006",
                "version": "0.200714.1807",
                "location": "sub-anm369962",
            },
        ),
        (
            "https://gui-beta-dandiarchive-org.netlify.app/#/dandiset/001001/"
            "draft/files?location=sub-RAT123%2F",
            "dandi-api",
            "folder",
            {"dandiset_id": "001001", "version": "draft", "location": "sub-RAT123/"},
        ),
        # by direct instance name ad-hoc URI instance:ID[@version][/path]
        (
            "dandi://dandi-api-local-docker-tests/000002@draft",
            "dandi-api-local-docker-tests",
            "dandiset",
            {"dandiset_id": "000002", "version": "draft"},
        ),
        (
            "dandi://dandi-api-local-docker-tests/000002@draft/path",
            "dandi-api-local-docker-tests",
            "item",
            {"dandiset_id": "000002", "location": "path", "version": "draft"},
        ),
        (
            "dandi://dandi-api-local-docker-tests/000002/path",
            "dandi-api-local-docker-tests",
            "item",
            {"dandiset_id": "000002", "location": "path", "version": None},
        ),
        (  # test on "public" instance and have trailing / to signal the folder
            "dandi://dandi-api/000002/path/",
            "dandi-api",
            "folder",
            {"dandiset_id": "000002", "location": "path/", "version": None},
        ),
        # TODO: bring back a test on deploy-preview-
        # # And the hybrid for "drafts" where it still goes by girder ID
        # (
        #     "https://deploy-preview-341--gui-dandiarchive-org.netlify.app/#/dandiset/000027"
        #     "/draft/files?_id=5f176583f63d62e1dbd06943&_modelType=folder",
        #     "dandi-api",
        #     "folder",
        #     {
        #         "dandiset_id": "000027",
        #         "version": "draft",
        #         "folder_id": "5f176583f63d62e1dbd06943",
        #     },
        # )
        (
            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download/",
            "dandi-api",
            "item",
            {
                "dandiset_id": "000003",
                "version": "draft",
                "asset_id": "0a748f90-d497-4a9c-822e-9c63811db412",
            },
        ),
    ],
)
def test_parse_api_url(url, instance, asset_type, asset_id):
    st, s, a, aid = parse_dandi_url(url)
    assert st == "api"
    assert s == known_instances[instance].api + "/"
    assert a == asset_type
    assert aid == asset_id


@mark.skipif_no_network
def test_parse_dandi_url_not_found():
    # Unlikely this one would ever come to existence
    with pytest.raises(NotFoundError):
        parse_dandi_url("https://dandiarchive.org/dandiset/999999")


@mark.skipif_no_network
def test_follow_redirect():
    assert (
        follow_redirect("https://bit.ly/dandi12")
        == "https://gui.dandiarchive.org/#/file-browser/folder/5e72b6ac3da50caa9adb0498"
    )



import json
import os
import os.path as op
from pathlib import Path
from shutil import rmtree

import pytest
import tqdm

from ..download import download
from ..girder import TQDMProgressReporter
from ..utils import find_files


# both urls point to 000027 (lean test dataset), and both draft and "released"
# version have only a single file ATM
@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027(url, tmpdir):
    ret = download(url, tmpdir)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == [
        "dandiset.yaml",
        "sub-RAT123",
        op.join("sub-RAT123", "sub-RAT123.nwb"),
    ]
    # and checksum should be correct as well
    from ..support.digests import Digester

    assert (
        Digester(["md5"])(dsdir / "sub-RAT123" / "sub-RAT123.nwb")["md5"]
        == "33318fd510094e4304868b4a481d4a5a"
    )
    # redownload - since already exist there should be an exception
    with pytest.raises(FileExistsError):
        download(url, tmpdir)

    # TODO: somehow get that status report about what was downloaded and what not
    download(url, tmpdir, existing="skip")  # TODO: check that skipped
    download(url, tmpdir, existing="overwrite")  # TODO: check that redownloaded
    download(url, tmpdir, existing="refresh")  # TODO: check that skipped (the same)


@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_metadata_only(url, tmpdir):
    ret = download(url, tmpdir, get_assets=False)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == ["dandiset.yaml"]


@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        pytest.param(
            "https://dandiarchive.org/dandiset/000027/0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_assets_only(url, tmpdir):
    ret = download(url, tmpdir, get_metadata=False)
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmpdir / "000027"
    downloads = (x.relto(dsdir) for x in dsdir.visit())
    assert sorted(downloads) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]


def test_girder_tqdm(monkeypatch):
    # smoke test to ensure we do not blow up
    def raise_assertion_error(*args, **kwargs):
        assert False, "pretend locking failed"

    monkeypatch.setattr(tqdm, "tqdm", raise_assertion_error)

    with TQDMProgressReporter() as pr:
        pr.update(10)


@pytest.mark.parametrize("resizer", [lambda sz: 0, lambda sz: sz // 2, lambda sz: sz])
@pytest.mark.parametrize(
    "version",
    [
        pytest.param(
            "0.200721.2222",
            marks=pytest.mark.xfail(reason="publish.dandiarchive.org is gone"),
        ),
        "draft",
    ],
)
def test_download_000027_resume(tmp_path, resizer, version):
    from ..support.digests import Digester

    url = f"https://dandiarchive.org/dandiset/000027/{version}"
    digester = Digester()
    download(url, tmp_path, get_metadata=False)
    dsdir = tmp_path / "000027"
    nwb = dsdir / "sub-RAT123" / "sub-RAT123.nwb"
    digests = digester(str(nwb))
    dldir = nwb.with_name(nwb.name + ".dandidownload")
    dldir.mkdir()
    dlfile = dldir / "file"
    nwb.rename(dlfile)
    size = dlfile.stat().st_size
    os.truncate(dlfile, resizer(size))
    with (dldir / "checksum").open("w") as fp:
        json.dump(digests, fp)
    download(url, tmp_path, get_metadata=False)
    contents = [
        op.relpath(op.join(dirpath, entry), dsdir)
        for (dirpath, dirnames, filenames) in os.walk(dsdir)
        for entry in dirnames + filenames
    ]
    assert sorted(contents) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]
    assert nwb.stat().st_size == size
    assert digester(str(nwb)) == digests


def test_download_newest_version(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}", tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"
    text_dandiset["client"].publish_version(dandiset_id, "draft")
    (text_dandiset["dspath"] / "file.txt").write_text("This is different text.\n")
    text_dandiset["reupload"]()
    rmtree(tmp_path / dandiset_id)
    download(f"{local_dandi_api['instance'].api}/dandisets/{dandiset_id}", tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"


def test_download_folder(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(
        f"dandi://{local_dandi_api['instance_id']}/{dandiset_id}/subdir2/", tmp_path
    )
    assert sorted(map(Path, find_files(r".*", paths=[tmp_path], dirs=True))) == [
        tmp_path / "subdir2",
        tmp_path / "subdir2" / "banana.txt",
        tmp_path / "subdir2" / "coconut.txt",
    ]
    assert (tmp_path / "subdir2" / "banana.txt").read_text() == "Banana\n"
    assert (tmp_path / "subdir2" / "coconut.txt").read_text() == "Coconut\n"


def test_download_item(local_dandi_api, text_dandiset, tmp_path):
    dandiset_id = text_dandiset["dandiset_id"]
    download(
        f"dandi://{local_dandi_api['instance_id']}/{dandiset_id}/subdir2/coconut.txt",
        tmp_path,
    )
    assert list(map(Path, find_files(r".*", paths=[tmp_path], dirs=True))) == [
        tmp_path / "coconut.txt"
    ]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"



from datetime import datetime

# PurePosixPath to be cast to for paths on girder
from pathlib import Path, PurePosixPath
import re
import sys
import time

from .consts import (
    collection_drafts,
    dandiset_identifier_regex,
    dandiset_metadata_file,
    metadata_digests,
)
from . import __version__, lgr
from .utils import ensure_datetime, ensure_strtime, get_instance


def upload(
    paths,
    existing="refresh",
    validation="require",
    dandiset_path=None,
    girder_collection=collection_drafts,
    girder_top_folder=None,
    dandi_instance="dandi",
    fake_data=False,  # TODO: not implemented, prune?
    allow_any_path=False,
    upload_dandiset_metadata=False,
    devel_debug=False,
    jobs=None,
    jobs_per_file=None,
):
    from .dandiset import Dandiset
    from . import girder
    from .support.digests import Digester

    dandiset = Dandiset.find(dandiset_path)
    if not dandiset:
        raise RuntimeError(
            f"Found no {dandiset_metadata_file} anywhere.  "
            "Use 'dandi register', 'download', or 'organize' first"
        )

    # Should no longer be needed
    # dandiset_path = Path(dandiset_path).resolve()

    instance = get_instance(dandi_instance)
    if instance.girder is None:
        assert instance.api is not None
        return _new_upload(
            instance.api,
            dandiset,
            paths,
            existing,
            validation,
            dandiset_path,
            allow_any_path,
            upload_dandiset_metadata,
            devel_debug,
            jobs=jobs,
            jobs_per_file=jobs_per_file,
        )

    if upload_dandiset_metadata:
        raise NotImplementedError(
            "Upload of dandiset metadata to Girder based server is not supported."
        )

    client = girder.get_client(instance.girder)

    # Girder side details:

    if not girder_collection:
        girder_collection = collection_drafts

    if not girder_top_folder:
        # We upload to staging/dandiset_id
        ds_identifier = dandiset.identifier
        if not ds_identifier:
            raise ValueError(
                "No 'identifier' set for the dandiset yet.  Use 'dandi register'"
            )
        if not re.match(dandiset_identifier_regex, ds_identifier):
            raise ValueError(
                f"Dandiset identifier {ds_identifier} does not follow expected "
                f"convention {dandiset_identifier_regex!r}.  Use "
                f"'dandi register' to get a legit identifier"
            )
        # this is a path not a girder id
        girder_top_folder = ds_identifier
    girder_top_folder = PurePosixPath(girder_top_folder)

    if str(girder_top_folder) in (".", "..", "", "/"):
        raise ValueError(
            f"Got folder {girder_top_folder}, but files cannot be uploaded "
            f"into a collection directly."
        )

    import multiprocessing

    from .metadata import get_metadata
    from .pynwb_utils import get_object_id, ignore_benign_pynwb_warnings
    from .support.generatorify import generator_from_callback
    from .support.pyout import naturalsize
    from .utils import find_dandi_files, find_files, path_is_subpath
    from .validate import validate_file

    ignore_benign_pynwb_warnings()  # so validate doesn't whine

    try:
        collection_rec = girder.ensure_collection(client, girder_collection)
    except girder.gcl.HttpError as exc:
        if devel_debug:
            raise
        # provide a bit less intimidating error reporting
        lgr.error(
            "Failed to assure presence of the %s collection: %s",
            girder_collection,
            (girder.get_HttpError_response(exc) or {}).get("message", str(exc)),
        )
        sys.exit(1)

    lgr.debug("Working with collection %s", collection_rec)

    try:
        girder.lookup(client, girder_collection, path=girder_top_folder)
    except girder.GirderNotFound:
        raise ValueError(
            f"There is no {girder_top_folder} in {girder_collection}. "
            f"Did you use 'dandi register'?"
        )

    #
    # Treat paths
    #
    if not paths:
        paths = [dandiset.path]

    # Expand and validate all paths -- they should reside within dandiset
    paths = find_files(".*", paths) if allow_any_path else find_dandi_files(paths)
    paths = list(map(Path, paths))
    npaths = len(paths)
    lgr.info(f"Found {npaths} files to consider")
    for path in paths:
        if not (
            allow_any_path
            or path.name == dandiset_metadata_file
            or path.name.endswith(".nwb")
        ):
            raise NotImplementedError(
                f"ATM only .nwb and dandiset.yaml should be in the paths to upload. Got {path}"
            )
        if not path_is_subpath(str(path.absolute()), dandiset.path):
            raise ValueError(f"{path} is not under {dandiset.path}")

    # We will keep a shared set of "being processed" paths so
    # we could limit the number of them until
    #   https://github.com/pyout/pyout/issues/87
    # properly addressed
    process_paths = set()
    from collections import defaultdict

    uploaded_paths = defaultdict(lambda: {"size": 0, "errors": []})

    def skip_file(msg):
        return {"status": "skipped", "message": str(msg)}

    lock = multiprocessing.Lock()

    # TODO: we might want to always yield a full record so no field is not
    # provided to pyout to cause it to halt
    def process_path(path, relpath):
        """

        Parameters
        ----------
        path: Path
          Non Pure (OS specific) Path
        relpath:
          For location on Girder.  Will be cast to PurePosixPath

        Yields
        ------
        dict
          Records for pyout
        """
        # Ensure consistent types
        path = Path(path)
        relpath = PurePosixPath(relpath)
        try:
            try:
                path_stat = path.stat()
                yield {"size": path_stat.st_size}
            except FileNotFoundError:
                yield skip_file("ERROR: File not found")
                return
            except Exception as exc:
                # without limiting [:50] it might cause some pyout indigestion
                yield skip_file("ERROR: %s" % str(exc)[:50])
                return

            yield {"status": "checking girder"}

            girder_folder = girder_top_folder / relpath.parent

            # we will add some fields which would help us with deciding to
            # reupload or not
            file_metadata_ = {
                "uploaded_size": path_stat.st_size,
                "uploaded_mtime": ensure_strtime(path_stat.st_mtime),
                # "uploaded_date": None,  # to be filled out upon upload completion
            }

            # A girder delete API target to .delete before uploading a file
            # (e.g. if decided to reupload)
            delete_before_upload = None

            def ensure_item():
                """This function might need to be called twice, e.g. if we
                are to reupload the entire item.

                ATM new versions of the files would create new items since
                the policy is one File per Item
                """
                try:
                    lock.acquire(timeout=60)
                    # TODO: we need to make this all thread safe all the way
                    #       until uploading the file since multiple threads would
                    #       create multiple
                    # ATM it even fails with  No such folder: 5e33658d6eb14e0bf49e97d5",
                    # so will first upload one file and then the rest... not sure why
                    # locking doesn't work
                    folder_rec = girder.ensure_folder(
                        client, collection_rec, girder_collection, girder_folder
                    )

                    # Get (if already exists) or create an item
                    item_rec = client.createItem(
                        folder_rec["_id"], name=relpath.name, reuseExisting=True
                    )
                finally:
                    lock.release()
                return item_rec

            def ensure_folder():
                try:
                    lock.acquire(timeout=60)
                    folder_rec = girder.ensure_folder(
                        client, collection_rec, girder_collection, girder_folder
                    )
                finally:
                    lock.release()
                return folder_rec

            #
            # 1. Validate first, so we do not bother girder at all if not kosher
            #
            # TODO: enable back validation of dandiset.yaml
            if path.name != dandiset_metadata_file and validation != "skip":
                yield {"status": "pre-validating"}
                validation_errors = validate_file(path)
                yield {"errors": len(validation_errors)}
                # TODO: split for dandi, pynwb errors
                if validation_errors:
                    if validation == "require":
                        yield skip_file("failed validation")
                        return
                else:
                    yield {"status": "validated"}
            else:
                # yielding empty causes pyout to get stuck or crash
                # https://github.com/pyout/pyout/issues/91
                # yield {"errors": '',}
                pass

            #
            # Special handling for dandiset.yaml
            # Yarik hates it but that is life for now. TODO
            #
            if path.name == dandiset_metadata_file:
                # TODO This is a temporary measure to avoid breaking web UI
                # dandiset metadata schema assumptions.  All edits should happen
                # online.
                yield skip_file("should be edited online")
                return
                # We need to upload its content as metadata for the entire
                # folder.
                folder_rec = ensure_folder()
                remote_metadata = folder_rec["meta"]
                if remote_metadata.get("dandiset", {}) == dandiset.metadata:
                    yield skip_file("exists (same)")
                else:
                    remote_metadata["dandiset"] = dandiset.metadata
                    yield {"status": "uploading dandiset metadata"}
                    client.addMetadataToFolder(folder_rec["_id"], remote_metadata)
                    yield {"status": "done"}
                # Interrupt -- no file to upload
                return

            #
            # 2. Ensure having an item
            #
            item_rec = ensure_item()

            #
            # 3. Analyze possibly present on the remote files in the item
            #
            file_recs = list(client.listFile(item_rec["_id"]))

            # get metadata and if we have all indications that it is
            # probably the same -- we just skip
            stat_fields = [
                # Care only about mtime, ignore ctime which could change
                "uploaded_mtime",
                "uploaded_size",
            ]
            assert sorted(file_metadata_) == stat_fields
            item_file_metadata_ = {
                k: item_rec.get("meta", {}).get(k, None) for k in stat_fields
            }
            lgr.debug(
                "Files meta: local file: %s  remote file: %s",
                file_metadata_,
                item_file_metadata_,
            )

            if item_file_metadata_["uploaded_mtime"]:
                local_mtime = ensure_datetime(file_metadata_["uploaded_mtime"])
                remote_mtime = ensure_datetime(
                    item_file_metadata_.get("uploaded_mtime")
                )
                remote_file_status = (
                    "same"
                    if (file_metadata_ == item_file_metadata_)
                    else (
                        "newer"
                        if remote_mtime > local_mtime
                        else ("older" if remote_mtime < local_mtime else "diff")
                    )
                )
            else:
                remote_file_status = "no mtime"
            exists_msg = f"exists ({remote_file_status})"

            if file_recs:  # there is a file already
                if len(file_recs) > 1:
                    lgr.debug(f"Item {item_rec} contains multiple files: {file_recs}")
                if existing == "error":
                    # as promised -- not gentle at all!
                    raise FileExistsError(exists_msg)
                if existing == "skip":
                    yield skip_file(exists_msg)
                    return
                # Logic below only for overwrite and reupload
                if existing == "overwrite":
                    if remote_file_status == "same":
                        yield skip_file(exists_msg)
                        return
                elif existing == "refresh":
                    if not remote_file_status == "older":
                        yield skip_file(exists_msg)
                        return
                elif existing == "force":
                    pass
                else:
                    raise ValueError("existing")

                delete_before_upload = f'/item/{item_rec["_id"]}'

                yield {"message": exists_msg + " - reuploading"}

            #
            # 4. Extract metadata - delayed since takes time, but is done
            #    before actual upload, so we could skip if this fails
            #
            # Extract metadata before actual upload and skip if fails
            # TODO: allow for for non-nwb files to skip this step
            # ad-hoc for dandiset.yaml for now
            if path.name != dandiset_metadata_file:
                yield {"status": "extracting metadata"}
                try:
                    metadata = get_metadata(path)
                except Exception as exc:
                    if allow_any_path:
                        yield {"status": "failed to extract metadata"}
                        metadata = {}
                    else:
                        yield skip_file("failed to extract metadata: %s" % str(exc))
                        if not file_recs:
                            # remove empty item
                            yield {"status": "deleting empty item"}
                            client.delete(f'/item/{item_rec["_id"]}')
                            yield {"status": "deleted empty item"}
                        return

            #
            # ?. Compute checksums and possible other digests (e.g. for s3, ipfs - TODO)
            #
            yield {"status": "digesting"}
            try:
                # TODO: in theory we could also cache the result, but since it is
                # critical to get correct checksums, safer to just do it all the time.
                # Should typically be faster than upload itself ;-)
                digester = Digester(metadata_digests)
                file_metadata_.update(digester(path))
            except Exception as exc:
                yield skip_file("failed to compute digests: %s" % str(exc))
                return

            #
            # 5. Upload file
            #
            # TODO: we could potentially keep new item "hidden" until we are
            #  done with upload, and only then remove old one and replace with
            #  a new one (rename from "hidden" name).
            if delete_before_upload:
                yield {"status": "deleting old"}
                client.delete(delete_before_upload)
                yield {"status": "old deleted"}
                # create a a new item
                item_rec = ensure_item()

            yield {"status": "uploading"}
            # Upload file to an item
            # XXX TODO progress reporting back to pyout is actually tricky
            #     if possible to implement via callback since
            #     callback would need to yield somehow from the context here.
            #     yoh doesn't see how that could be done yet. In the worst
            #     case we would copy uploadFileToItem and _uploadContents
            #     and make them into generators to relay progress instead of
            #     via callback
            # https://stackoverflow.com/questions/9968592/turn-functions-with-a-callback-into-python-generators
            # has some solutions but all IMHO are abit too complex

            for r in generator_from_callback(
                lambda c: client.uploadFileToItem(
                    item_rec["_id"], str(path), progressCallback=c
                )
            ):
                upload_perc = 100 * ((r["current"] / r["total"]) if r["total"] else 1.0)
                if girder._DANDI_LOG_GIRDER:
                    girder.lgr.debug(
                        "PROGRESS[%s]: done=%d %%done=%s",
                        str(path),
                        r["current"],
                        upload_perc,
                    )
                uploaded_paths[str(path)]["size"] = r["current"]
                yield {"upload": upload_perc}

            # Get uploaded file id
            file_id, current = client.isFileCurrent(
                item_rec["_id"], path.name, path.absolute()
            )
            if not current:
                yield skip_file("File on server was unexpectedly changed")
                return

            # Compare file size against what download headers report
            # S3 doesn't seem to allow HEAD requests, so we need to instead do
            # a GET with a streaming response and not read the body.
            with client.sendRestRequest(
                "GET", f"file/{file_id}/download", jsonResp=False, stream=True
            ) as r:
                if int(r.headers["Content-Length"]) != path.stat().st_size:
                    yield skip_file("File size on server does not match local file")
                    return

            #
            # 6. Upload metadata
            #
            metadata_ = {}
            for k, v in metadata.items():
                if v in ("", None):
                    continue  # degenerate, why bother
                # XXX TODO: remove this -- it is only temporary, search should handle
                if isinstance(v, str):
                    metadata_[k] = v.lower()
                elif isinstance(v, datetime):
                    metadata_[k] = ensure_strtime(v)
            # we will add some fields which would help us with deciding to
            # reupload or not
            # .isoformat() would give is8601 representation but I see in girder
            # already
            # session_start_time   1971-01-01 12:00:00+00:00
            # decided to go for .isoformat for internal consistency -- let's see
            file_metadata_["uploaded_datetime"] = ensure_strtime(time.time())
            metadata_.update(file_metadata_)
            metadata_["uploaded_size"] = path_stat.st_size
            metadata_["uploaded_mtime"] = ensure_strtime(path_stat.st_mtime)
            metadata_["uploaded_by"] = "dandi %s" % __version__
            # Also store object_id for the file to help identify changes/moves
            try:
                metadata_["uploaded_nwb_object_id"] = get_object_id(str(path))
            except Exception as exc:
                (lgr.debug if allow_any_path else lgr.warning)(
                    "Failed to read object_id: %s", exc
                )

            # #
            # # 7. Also set remote file ctime to match local mtime
            # #   since for type "file", Resource has no "updated" field.
            # #   and this could us help to identify changes being done
            # #   to the remote file -- if metadata["uploaded_mtime"]
            # #   differs
            # yield {"status": "setting remote file timestamp"}
            # try:
            #     client.setResourceTimestamp(
            #         file_id, type="file", created=metadata_["uploaded_mtime"]
            #     )
            # except girder.gcl.HttpError as exc:
            #     if devel_debug:
            #         raise
            #     response = girder.get_HttpError_response(exc)
            #     message = response.get("message", str(exc))
            #     yield {"status": "WARNING", "message": message}

            # 7. Upload metadata
            yield {"status": "uploading metadata"}
            client.addMetadataToItem(item_rec["_id"], metadata_)
            yield {"status": "done"}

        except Exception as exc:
            if devel_debug:
                raise
            # Custom formatting for some exceptions we know to extract
            # user-meaningful message
            message = str(exc)
            if isinstance(exc, girder.gcl.HttpError):
                response = girder.get_HttpError_response(exc)
                if "message" in response:
                    message = response["message"]
            uploaded_paths[str(path)]["errors"].append(message)
            yield {"status": "ERROR", "message": message}
        finally:
            process_paths.remove(str(path))

    # We will again use pyout to provide a neat table summarizing our progress
    # with upload etc
    from .support import pyout as pyouts

    # for the upload speeds we need to provide a custom  aggregate
    t0 = time.time()

    def upload_agg(*ignored):
        dt = time.time() - t0
        total = sum(v["size"] for v in uploaded_paths.values())
        if not total:
            return ""
        speed = total / dt if dt else 0
        return "%s/s" % naturalsize(speed)

    pyout_style = pyouts.get_style(hide_if_missing=False)
    pyout_style["upload"]["aggregate"] = upload_agg

    rec_fields = ["path", "size", "errors", "upload", "status", "message"]
    out = pyouts.LogSafeTabular(style=pyout_style, columns=rec_fields)

    with out, client.lock_dandiset(dandiset.identifier):
        for path in paths:
            while len(process_paths) >= 10:
                lgr.log(2, "Sleep waiting for some paths to finish processing")
                time.sleep(0.5)

            rec = {"path": str(path)}
            process_paths.add(str(path))

            try:
                relpath = path.absolute().relative_to(dandiset.path)

                rec["path"] = str(relpath)
                if devel_debug:
                    # DEBUG: do serially
                    for v in process_path(path, relpath):
                        print(str(v), flush=True)
                else:
                    rec[tuple(rec_fields[1:])] = process_path(path, relpath)
            except ValueError as exc:
                if "does not start with" in str(exc):
                    # if top_path is not the top path for the path
                    # Provide more concise specific message without path details
                    rec.update(skip_file("must be a child of top path"))
                else:
                    rec.update(skip_file(exc))
            out(rec)

    # # Provide summary of errors if any recorded
    # # well -- they are also in summary by pyout.  So, for now - do not bother
    # # It would be worthwhile if we decide to log also other errors
    # errors = defaultdict(set)
    # for p, v in uploaded_paths.items():
    #     for e in v['errors']:
    #         errors[e].add(p)
    # if errors:
    #     lgr.error("Following errors were detected while uploading")
    #     for e, paths in errors.items():
    #         lgr.error(" %s: %d paths", e, len(paths))


def _new_upload(
    api_url,
    dandiset,
    paths,
    existing,
    validation,
    dandiset_path,
    allow_any_path,
    upload_dandiset_metadata,
    devel_debug,
    jobs=None,
    jobs_per_file=None,
):
    from .dandiapi import DandiAPIClient
    from .dandiset import APIDandiset
    from .support.digests import get_digest

    client = DandiAPIClient(api_url)
    client.dandi_authenticate()

    dandiset = APIDandiset(dandiset.path)  # "cast" to a new API based dandiset

    ds_identifier = dandiset.identifier
    # this is a path not a girder id

    if not re.match(dandiset_identifier_regex, str(ds_identifier)):
        raise ValueError(
            f"Dandiset identifier {ds_identifier} does not follow expected "
            f"convention {dandiset_identifier_regex!r}.  Use "
            f"'dandi register' to get a legit identifier"
        )

    from .metadata import get_default_metadata, nwb2asset
    from .pynwb_utils import ignore_benign_pynwb_warnings
    from .support.pyout import naturalsize
    from .utils import find_dandi_files, find_files, path_is_subpath
    from .validate import validate_file

    ignore_benign_pynwb_warnings()  # so validate doesn't whine

    #
    # Treat paths
    #
    if not paths:
        paths = [dandiset.path]

    # Expand and validate all paths -- they should reside within dandiset
    paths = find_files(".*", paths) if allow_any_path else find_dandi_files(paths)
    paths = list(map(Path, paths))
    npaths = len(paths)
    lgr.info(f"Found {npaths} files to consider")
    for path in paths:
        if not (
            allow_any_path
            or path.name == dandiset_metadata_file
            or path.name.endswith(".nwb")
        ):
            raise NotImplementedError(
                f"ATM only .nwb and dandiset.yaml should be in the paths to upload. Got {path}"
            )
        if not path_is_subpath(str(path.absolute()), dandiset.path):
            raise ValueError(f"{path} is not under {dandiset.path}")

    # We will keep a shared set of "being processed" paths so
    # we could limit the number of them until
    #   https://github.com/pyout/pyout/issues/87
    # properly addressed
    process_paths = set()
    from collections import defaultdict

    uploaded_paths = defaultdict(lambda: {"size": 0, "errors": []})

    def skip_file(msg):
        return {"status": "skipped", "message": str(msg)}

    # TODO: we might want to always yield a full record so no field is not
    # provided to pyout to cause it to halt
    def process_path(path, relpath):
        """

        Parameters
        ----------
        path: Path
          Non Pure (OS specific) Path
        relpath:
          For location on server.  Will be cast to PurePosixPath

        Yields
        ------
        dict
          Records for pyout
        """
        # Ensure consistent types
        path = Path(path)
        relpath = PurePosixPath(relpath)
        try:
            try:
                path_stat = path.stat()
                yield {"size": path_stat.st_size}
            except FileNotFoundError:
                yield skip_file("ERROR: File not found")
                return
            except Exception as exc:
                # without limiting [:50] it might cause some pyout indigestion
                yield skip_file("ERROR: %s" % str(exc)[:50])
                return

            #
            # Validate first, so we do not bother server at all if not kosher
            #
            # TODO: enable back validation of dandiset.yaml
            if path.name != dandiset_metadata_file and validation != "skip":
                yield {"status": "pre-validating"}
                validation_errors = validate_file(path)
                yield {"errors": len(validation_errors)}
                # TODO: split for dandi, pynwb errors
                if validation_errors:
                    if validation == "require":
                        yield skip_file("failed validation")
                        return
                else:
                    yield {"status": "validated"}
            else:
                # yielding empty causes pyout to get stuck or crash
                # https://github.com/pyout/pyout/issues/91
                # yield {"errors": '',}
                pass

            #
            # Special handling for dandiset.yaml
            # Yarik hates it but that is life for now. TODO
            #
            if path.name == dandiset_metadata_file:
                # TODO This is a temporary measure to avoid breaking web UI
                # dandiset metadata schema assumptions.  All edits should happen
                # online.
                if upload_dandiset_metadata:
                    yield {"status": "updating metadata"}
                    client.set_dandiset_metadata(
                        dandiset.identifier, metadata=dandiset.metadata
                    )
                    yield {"status": "updated metadata"}
                else:
                    yield skip_file("should be edited online")
                return

            #
            # Compute checksums
            #
            yield {"status": "digesting"}
            try:
                file_etag = get_digest(path, digest="dandi-etag")
            except Exception as exc:
                yield skip_file("failed to compute digest: %s" % str(exc))
                return

            extant = client.get_asset_bypath(ds_identifier, "draft", str(relpath))
            if extant is not None:
                # The endpoint used to search by paths doesn't include asset
                # metadata, so we need to make another API call:
                metadata = client.get_asset(ds_identifier, "draft", extant["asset_id"])
                local_mtime = ensure_datetime(path_stat.st_mtime)
                remote_mtime_str = metadata.get("blobDateModified")
                d = metadata.get("digest", {})
                if "dandi:dandi-etag" in d:
                    extant_etag = d["dandi:dandi-etag"]
                else:
                    # TODO: Should this error instead?
                    extant_etag = None
                if remote_mtime_str is not None:
                    remote_mtime = ensure_datetime(remote_mtime_str)
                    remote_file_status = (
                        "same"
                        if extant_etag == file_etag and remote_mtime == local_mtime
                        else (
                            "newer"
                            if remote_mtime > local_mtime
                            else ("older" if remote_mtime < local_mtime else "diff")
                        )
                    )
                else:
                    remote_mtime = None
                    remote_file_status = "no mtime"

                exists_msg = f"exists ({remote_file_status})"

                if existing == "error":
                    # as promised -- not gentle at all!
                    raise FileExistsError(exists_msg)
                if existing == "skip":
                    yield skip_file(exists_msg)
                    return
                # Logic below only for overwrite and reupload
                if existing == "overwrite":
                    if extant_etag == file_etag:
                        yield skip_file(exists_msg)
                        return
                elif existing == "refresh":
                    if extant_etag == file_etag:
                        yield skip_file("file exists")
                        return
                    elif remote_mtime is not None and remote_mtime >= local_mtime:
                        yield skip_file(exists_msg)
                        return
                elif existing == "force":
                    pass
                else:
                    raise ValueError(f"invalid value for 'existing': {existing!r}")

                yield {"message": f"{exists_msg} - reuploading"}

            #
            # Extract metadata - delayed since takes time, but is done before
            # actual upload, so we could skip if this fails
            #
            # Extract metadata before actual upload and skip if fails
            # TODO: allow for for non-nwb files to skip this step
            # ad-hoc for dandiset.yaml for now
            yield {"status": "extracting metadata"}
            try:
                asset_metadata = nwb2asset(
                    path, digest=file_etag, digest_type="dandi_etag"
                )
            except Exception as exc:
                if allow_any_path:
                    yield {"status": "failed to extract metadata"}
                    asset_metadata = get_default_metadata(
                        path, digest=file_etag, digest_type="dandi_etag"
                    )
                else:
                    yield skip_file("failed to extract metadata: %s" % str(exc))
                    return
            metadata = asset_metadata.json_dict()
            metadata["path"] = str(relpath)

            #
            # Upload file
            #
            yield {"status": "uploading"}
            validating = False
            for r in client.iter_upload(
                ds_identifier, "draft", metadata, str(path), jobs=jobs_per_file
            ):
                if r["status"] == "uploading":
                    uploaded_paths[str(path)]["size"] = r.pop("current")
                    yield r
                elif r["status"] == "post-validating":
                    # Only yield the first "post-validating" status
                    if not validating:
                        yield r
                        validating = True
                else:
                    yield r
            yield {"status": "done"}

        except Exception as exc:
            if devel_debug:
                raise
            # Custom formatting for some exceptions we know to extract
            # user-meaningful message
            message = str(exc)
            uploaded_paths[str(path)]["errors"].append(message)
            yield {"status": "ERROR", "message": message}
        finally:
            process_paths.remove(str(path))

    # We will again use pyout to provide a neat table summarizing our progress
    # with upload etc
    from .support import pyout as pyouts

    # for the upload speeds we need to provide a custom  aggregate
    t0 = time.time()

    def upload_agg(*ignored):
        dt = time.time() - t0
        # to help avoiding dict length changes during upload
        # might be not a proper solution
        # see https://github.com/dandi/dandi-cli/issues/502 for more info
        uploaded_recs = list(uploaded_paths.values())
        total = sum(v["size"] for v in uploaded_recs)
        if not total:
            return ""
        speed = total / dt if dt else 0
        return "%s/s" % naturalsize(speed)

    pyout_style = pyouts.get_style(hide_if_missing=False)
    pyout_style["upload"]["aggregate"] = upload_agg

    rec_fields = ["path", "size", "errors", "upload", "status", "message"]
    out = pyouts.LogSafeTabular(style=pyout_style, columns=rec_fields, max_workers=jobs)

    with out, client.session():
        for path in paths:
            while len(process_paths) >= 10:
                lgr.log(2, "Sleep waiting for some paths to finish processing")
                time.sleep(0.5)

            rec = {"path": str(path)}
            process_paths.add(str(path))

            try:
                relpath = path.absolute().relative_to(dandiset.path)

                rec["path"] = str(relpath)
                if devel_debug:
                    # DEBUG: do serially
                    for v in process_path(path, relpath):
                        print(str(v), flush=True)
                else:
                    rec[tuple(rec_fields[1:])] = process_path(path, relpath)
            except ValueError as exc:
                if "does not start with" in str(exc):
                    # if top_path is not the top path for the path
                    # Provide more concise specific message without path details
                    rec.update(skip_file("must be a child of top path"))
                else:
                    rec.update(skip_file(exc))
            out(rec)



import re

import pytest

from ..consts import dandiset_identifier_regex, dandiset_metadata_file
from ..register import register
from ..utils import yaml_load


@pytest.mark.parametrize("present", [True, False])
def test_smoke_register(local_docker_compose_env, tmp_path, present):
    if present:
        (tmp_path / dandiset_metadata_file).write_text("{}\n")
    dandiset_metadata = register(
        "Dandiset Name",
        "Dandiset Description",
        dandiset_path=tmp_path,
        dandi_instance=local_docker_compose_env["instance_id"],
    )
    assert dandiset_metadata
    assert dandiset_metadata["name"] == "Dandiset Name"
    assert dandiset_metadata["description"] == "Dandiset Description"
    assert re.match(dandiset_identifier_regex, dandiset_metadata["identifier"])

    # Verify that dandiset.yaml was updated or generated
    with (tmp_path / dandiset_metadata_file).open() as fp:
        metadata = yaml_load(fp, typ="base")
    assert metadata == dandiset_metadata

    # TODO: Check that a Dandiset exists in the local-docker-tests instance
    # with the given identifier



import os
import os.path as op

from .consts import dandiset_metadata_file, routes
from .dandiset import Dandiset
from . import get_logger, girder
from .utils import get_instance

lgr = get_logger()


def register(name, description, dandiset_path=None, dandi_instance="dandi"):
    """Register a dandiset

    Parameters
    ----------
    name: str
    description: str
    dandiset_path: str, optional
    dandi_instance: str, optional

    Returns
    -------
    dict
      Metadata record of the registered dandiset
    """
    dandi_instance = get_instance(dandi_instance)
    if not dandiset_path and op.exists(dandiset_metadata_file):
        dandiset = Dandiset.find(os.getcwd())
        if dandiset:
            if "identifier" in dandiset.metadata:
                lgr.warning(
                    "Running 'register' while in a dandiset at %s.  We will "
                    "not generate %s",
                    dandiset,
                    dandiset_metadata_file,
                )
            else:
                dandiset_path = dandiset.path
                lgr.info(
                    "We will populate %s of the %s dandiset",
                    dandiset_metadata_file,
                    dandiset.path,
                )
        else:
            lgr.info(
                "No dandiset path was provided and no dandiset detected in the path."
                " No %s will be modified",
                dandiset_metadata_file,
            )

    client = girder.get_client(dandi_instance.girder)
    dandiset = client.register_dandiset(name, description)

    url = routes.dandiset_draft.format(**locals())

    lgr.info(
        f"Registered dandiset {dandiset['identifier']} at {url}. Please visit and adjust metadata."
    )

    if dandiset_path:
        lgr.info(f"Adjusting {dandiset_path} with obtained metadata")
        ds = Dandiset(dandiset_path, allow_empty=True)
        ds.update_metadata(dandiset)

    return dandiset



import click

from .base import devel_debug_option, instance_option, map_to_click_exceptions


@click.command()
@click.argument("paths", nargs=-1, type=click.Path(exists=False, dir_okay=True))
@instance_option()
@devel_debug_option()
@map_to_click_exceptions
def delete(paths, dandi_instance="dandi", devel_debug=False):
    """ Delete Dandisets and assets from the server """
    from ..delete import delete

    delete(paths, dandi_instance=dandi_instance, devel_debug=devel_debug)



import click

from .base import dandiset_path_option, instance_option, map_to_click_exceptions


@click.command()
@dandiset_path_option(
    help="Top directory (local) for the dandiset, where dandi will download "
    "(or update existing) dandiset.yaml upon successful registration."
)
@click.option(
    "-n", "--name", help="Short name or title for the dandiset.", prompt="Name"
)
@click.option(
    "-D",
    "--description",
    help="Description of the dandiset - high level summary of the experiment "
    "and type(s) of data.",
    prompt="Description",
)
# &
# Development options:  Set DANDI_DEVEL for them to become available
#
# TODO: should always go to dandi for now
@instance_option()
@map_to_click_exceptions
def register(name, description, dandiset_path=None, dandi_instance="dandi"):
    """Register a new dandiset in the DANDI archive.

    This command provides only a minimal set of metadata. It is
    recommended to use Web UI to fill out other metadata fields for the
    dandiset
    """
    from ..register import register

    dandiset = register(name, description, dandiset_path, dandi_instance)
    print("identifier:", dandiset.identifier)
